# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import (
    FloatTensorType, Int64TensorType, DoubleTensorType)
from ..common.utils import (
    check_input_and_output_numbers, check_input_and_output_types)


def calculate_pls_regression_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[
            FloatTensorType, Int64TensorType, DoubleTensorType])

    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError('Input must be a [N, C]-tensor')

    op = operator.raw_operator
    cls_type = operator.inputs[0].type.__class__
    if cls_type != DoubleTensorType:
        cls_type = FloatTensorType
    N = operator.inputs[0].get_first_dimension()
    operator.outputs[0].type = cls_type([N, op.coef_.shape[1]])


register_shape_calculator('SklearnPLSRegression',
                          calculate_pls_regression_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from sklearn.base import is_regressor
from sklearn.svm import LinearSVC
from ..proto import onnx_proto
from ..common._apply_operation import (
    apply_concat, apply_identity, apply_mul, apply_reshape)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_normalization
from ..common._apply_operation import (
    apply_slice, apply_sub, apply_cast, apply_abs, apply_add, apply_div)
from ..common.utils_classifier import _finalize_converter_classes
from ..common.data_types import guess_proto_type, Int64TensorType
from .._supported_operators import sklearn_operator_name_map


def _iteration_one_versus(scope, container, inputs, i, estimator, cl_type,
                          proto_dtype, use_raw_scores=True, prob_shape=None):
    op_type = sklearn_operator_name_map[type(estimator)]

    this_operator = scope.declare_local_operator(
        op_type, raw_model=estimator)
    this_operator.inputs = inputs

    if is_regressor(estimator):
        score_name = scope.declare_local_variable('score_%d' % i, cl_type())
        this_operator.outputs.append(score_name)

        if hasattr(estimator, 'coef_') and len(estimator.coef_.shape) == 2:
            raise RuntimeError(
                "OneVsRestClassifier or OneVsOneClassifier accepts "
                "regressor with only one target.")
        p1 = score_name.onnx_name
    else:
        if container.has_options(estimator, 'raw_scores'):
            container.add_options(
                id(estimator), {'raw_scores': use_raw_scores})
            scope.add_options(
                id(estimator), {'raw_scores': use_raw_scores})
        label_name = scope.declare_local_variable(
            'label_%d' % i, Int64TensorType())
        prob_name = scope.declare_local_variable('proba_%d' % i, cl_type())
        this_operator.outputs.append(label_name)
        this_operator.outputs.append(prob_name)

        # gets the probability for the class 1
        p1 = scope.get_unique_variable_name('probY_%d' % i)
        if isinstance(estimator, LinearSVC):
            apply_identity(scope, prob_name.onnx_name, p1, container)
        else:
            apply_slice(scope, prob_name.onnx_name, p1, container, starts=[1],
                        ends=[2], axes=[1])
    return None, None, p1


def convert_one_vs_rest_classifier(scope: Scope, operator: Operator,
                                   container: ModelComponentContainer):
    """
    Converts a *OneVsRestClassifier* into *ONNX* format.
    """
    if scope.get_options(operator.raw_operator, dict(nocl=False))['nocl']:
        raise RuntimeError(
            "Option 'nocl' is not implemented for operator '{}'.".format(
                operator.raw_operator.__class__.__name__))
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator
    options = container.get_options(op, dict(raw_scores=False))
    use_raw_scores = options['raw_scores']
    probs_names = []
    cl_type = operator.inputs[0].type.__class__
    prob_shape = None
    for i, estimator in enumerate(op.estimators_):
        prob_shape, _, p1 = _iteration_one_versus(
            scope, container, operator.inputs, i, estimator, cl_type,
            proto_dtype, use_raw_scores, prob_shape=prob_shape)
        probs_names.append(p1)

    if op.multilabel_:
        # concatenates outputs
        conc_name = operator.outputs[1].full_name
        apply_concat(scope, probs_names, conc_name, container, axis=1)

        # builds the labels (matrix with integer)
        # scikit-learn may use probabilities or raw score
        # but ONNX converters only uses probabilities.
        # https://github.com/scikit-learn/scikit-learn/sklearn/
        # multiclass.py#L290
        # Raw score would mean: scores = conc_name.
        thresh_name = scope.get_unique_variable_name('thresh')
        container.add_initializer(
            thresh_name, proto_dtype,
            [1, len(op.classes_)], [.5] * len(op.classes_))
        scores = scope.get_unique_variable_name('threshed')
        apply_sub(scope, [conc_name, thresh_name], scores, container)

        # sign
        signed_input = scope.get_unique_variable_name('signed')
        container.add_node('Sign', [scores], [signed_input],
                           name=scope.get_unique_operator_name('Sign'))
        # clip
        signed_input_cast = scope.get_unique_variable_name('signed_int64')
        apply_cast(scope, signed_input, signed_input_cast,
                   container, to=onnx_proto.TensorProto.INT64)

        label_name = scope.get_unique_variable_name('label')
        if container.target_opset <= 11:
            abs_name = scope.get_unique_variable_name('abs')
            add_name = scope.get_unique_variable_name('add')
            cst_2 = scope.get_unique_variable_name('cst2')
            container.add_initializer(
                cst_2, onnx_proto.TensorProto.INT64, [1], [2])
            apply_abs(scope, [signed_input_cast], [abs_name], container)
            apply_add(scope, [signed_input_cast, abs_name], [add_name],
                      container)
            apply_div(
                scope, [add_name, cst_2], [label_name],
                container)
        else:
            zero_cst = scope.get_unique_variable_name('zero')
            container.add_initializer(
                zero_cst, onnx_proto.TensorProto.INT64, [], [0])
            container.add_node(
                'Clip', [signed_input_cast, zero_cst],
                [label_name],
                name=scope.get_unique_operator_name('Clip'))
        apply_reshape(scope, [label_name], [operator.outputs[0].full_name],
                      container, desired_shape=(-1, op.n_classes_))
    else:
        # concatenates outputs
        conc_name = scope.get_unique_variable_name('concatenated')
        apply_concat(scope, probs_names, conc_name, container, axis=1)
        if len(op.estimators_) == 1:
            zeroth_col_name = scope.get_unique_variable_name('zeroth_col')
            merged_prob_name = scope.get_unique_variable_name('merged_prob')
            unit_float_tensor_name = scope.get_unique_variable_name(
                'unit_float_tensor')
            if use_raw_scores:
                container.add_initializer(
                    unit_float_tensor_name, proto_dtype, [], [-1.0])
                apply_mul(scope, [unit_float_tensor_name, conc_name],
                          zeroth_col_name, container, broadcast=1)
            else:
                container.add_initializer(
                    unit_float_tensor_name, proto_dtype, [], [1.0])
                apply_sub(scope, [unit_float_tensor_name, conc_name],
                          zeroth_col_name, container, broadcast=1)
            apply_concat(scope, [zeroth_col_name, conc_name],
                         merged_prob_name, container, axis=1)
            conc_name = merged_prob_name

        if use_raw_scores:
            apply_identity(scope, conc_name,
                           operator.outputs[1].full_name, container)
        else:
            # normalizes the outputs
            apply_normalization(
                scope, conc_name, operator.outputs[1].full_name,
                container, axis=1, p=1)

        # extracts the labels
        label_name = scope.get_unique_variable_name('label_name')
        container.add_node('ArgMax', conc_name, label_name,
                           name=scope.get_unique_operator_name('ArgMax'),
                           axis=1)

        _finalize_converter_classes(scope, label_name,
                                    operator.outputs[0].full_name, container,
                                    op.classes_, proto_dtype)


register_converter('SklearnOneVsRestClassifier',
                   convert_one_vs_rest_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'raw_scores': [True, False]})



# SPDX-License-Identifier: Apache-2.0


from ..proto import onnx_proto
from ..common._apply_operation import (
    apply_cast, apply_div, apply_sqrt, apply_sub)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import (
    Int64TensorType, DoubleTensorType, FloatTensorType, guess_proto_type)


def convert_truncated_svd(scope: Scope, operator: Operator,
                          container: ModelComponentContainer):
    # Create alias for the scikit-learn truncated SVD model we
    # are going to convert
    svd = operator.raw_operator
    if isinstance(operator.inputs[0].type, DoubleTensorType):
        proto_dtype = guess_proto_type(operator.inputs[0].type)
    else:
        proto_dtype = guess_proto_type(FloatTensorType())
    # Transpose [K, C] matrix to [C, K], where C/K is the
    # input/transformed feature dimension
    transform_matrix = svd.components_.transpose()
    transform_matrix_name = scope.get_unique_variable_name('transform_matrix')
    # Put the transformation into an ONNX tensor
    container.add_initializer(
        transform_matrix_name, proto_dtype,
        transform_matrix.shape, transform_matrix.flatten())

    input_name = operator.inputs[0].full_name
    if isinstance(operator.inputs[0].type, Int64TensorType):
        cast_output_name = scope.get_unique_variable_name('cast_output')

        apply_cast(scope, input_name, cast_output_name, container,
                   to=onnx_proto.TensorProto.FLOAT)
        input_name = cast_output_name

    if operator.type == 'SklearnTruncatedSVD':
        # Create the major operator, a matrix multiplication.
        container.add_node(
            'MatMul', [input_name, transform_matrix_name],
            operator.outputs[0].full_name, name=operator.full_name)
    else:  # PCA
        if svd.mean_ is not None:
            mean_name = scope.get_unique_variable_name('mean')
            sub_result_name = scope.get_unique_variable_name('sub_result')

            container.add_initializer(mean_name, proto_dtype,
                                      svd.mean_.shape, svd.mean_)

            # Subtract mean from input tensor
            apply_sub(scope, [input_name, mean_name],
                      sub_result_name, container, broadcast=1)
        else:
            sub_result_name = input_name
        if svd.whiten:
            explained_variance_name = scope.get_unique_variable_name(
                'explained_variance')
            explained_variance_root_name = scope.get_unique_variable_name(
                'explained_variance_root')
            matmul_result_name = scope.get_unique_variable_name(
                'matmul_result')

            container.add_initializer(
                explained_variance_name, proto_dtype,
                svd.explained_variance_.shape, svd.explained_variance_)

            container.add_node(
                'MatMul', [sub_result_name, transform_matrix_name],
                matmul_result_name,
                name=scope.get_unique_operator_name('MatMul'))
            apply_sqrt(scope, explained_variance_name,
                       explained_variance_root_name, container)
            apply_div(scope,
                      [matmul_result_name, explained_variance_root_name],
                      operator.outputs[0].full_name, container, broadcast=1)
        else:
            container.add_node(
                'MatMul', [sub_result_name, transform_matrix_name],
                operator.outputs[0].full_name,
                name=scope.get_unique_operator_name('MatMul'))


register_converter('SklearnIncrementalPCA', convert_truncated_svd)
register_converter('SklearnPCA', convert_truncated_svd)
register_converter('SklearnTruncatedSVD', convert_truncated_svd)



# SPDX-License-Identifier: Apache-2.0

"""
Functions to calculate output shapes of linear classifiers
and regressors.
"""
import numbers
import numpy as np
from .data_types import (
    BooleanTensorType,
    DoubleTensorType,
    FloatTensorType,
    Int64TensorType,
    StringTensorType,
)
from .utils import check_input_and_output_numbers, check_input_and_output_types
from .utils_classifier import get_label_classes


def calculate_linear_classifier_output_shapes(operator):
    """
    This operator maps an input feature vector into a scalar label if
    the number of outputs is one. If two outputs appear in this
    operator's output list, we should further generate a tensor storing
    all classes' probabilities.

    Allowed input/output patterns are
        1. [N, C] ---> [N, 1], A sequence of map

    """
    _calculate_linear_classifier_output_shapes(operator)


def _calculate_linear_classifier_output_shapes(
        operator, decision_path=False, decision_leaf=False,
        enable_type_checking=True):
    n_out = 0
    if decision_path:
        n_out += 1
    if decision_leaf:
        n_out += 1
    out_range = [2, 2 + n_out]
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=out_range)
    if enable_type_checking:
        check_input_and_output_types(operator, good_input_types=[
            BooleanTensorType, DoubleTensorType,
            FloatTensorType, Int64TensorType])

    N = operator.inputs[0].get_first_dimension()
    op = operator.raw_operator
    class_labels = get_label_classes(operator.scope_inst, op)

    number_of_classes = len(class_labels)
    if all(isinstance(i, np.ndarray) for i in class_labels):
        class_labels = np.concatenate(class_labels)
    if all(isinstance(i, str) for i in class_labels):
        shape = ([N, len(op.classes_)]
                 if (getattr(op, 'multilabel_', False) or (
                     isinstance(op.classes_, list) and
                     isinstance(op.classes_[0], np.ndarray))) else [N])
        operator.outputs[0].set_type(StringTensorType(shape=shape))
        if number_of_classes > 2 or operator.type != 'SklearnLinearSVC':
            shape = ([len(op.classes_), N, max([len(x) for x in op.classes_])]
                     if isinstance(op.classes_, list)
                     and isinstance(op.classes_[0], np.ndarray)
                     else [N, number_of_classes])
            operator.outputs[1].type.shape = shape
        else:
            # For binary LinearSVC, we produce probability of
            # the positive class
            operator.outputs[1].type.shape = [N, 1]
    elif all(isinstance(i, (numbers.Real, bool, np.bool_))
             for i in class_labels):
        shape = ([N, len(op.classes_)]
                 if (getattr(op, 'multilabel_', False) or (
                     isinstance(op.classes_, list) and
                     isinstance(op.classes_[0], np.ndarray))) else [N])
        operator.outputs[0].set_type(Int64TensorType(shape=shape))
        if number_of_classes > 2 or operator.type != 'SklearnLinearSVC':
            shape = ([len(op.classes_), N, max([len(x) for x in op.classes_])]
                     if isinstance(op.classes_, list)
                     and isinstance(op.classes_[0], np.ndarray)
                     else [N, number_of_classes])
            operator.outputs[1].type.shape = shape
        else:
            # For binary LinearSVC, we produce probability of
            # the positive class
            operator.outputs[1].type.shape = [N, 1]
    else:
        raise ValueError('Label types must be all integers or all strings.')

    # decision_path, decision_leaf
    for n in range(2, len(operator.outputs)):
        operator.outputs[n].type.shape = [N, 1]


def calculate_linear_regressor_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C] ---> [N, 1]

    This operator produces a scalar prediction for every example in a
    batch. If the input batch size is N, the output shape may be
    [N, 1].
    """
    _calculate_linear_regressor_output_shapes(operator)


def _calculate_linear_regressor_output_shapes(
        operator, enable_type_checking=True):
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=1)
    if enable_type_checking:
        check_input_and_output_types(operator, good_input_types=[
            BooleanTensorType, DoubleTensorType,
            FloatTensorType, Int64TensorType])

    inp0 = operator.inputs[0].type
    if isinstance(inp0, (FloatTensorType, DoubleTensorType)):
        cls_type = inp0.__class__
    else:
        cls_type = FloatTensorType

    N = operator.inputs[0].get_first_dimension()
    if (hasattr(operator.raw_operator, 'coef_') and
            len(operator.raw_operator.coef_.shape) > 1):
        operator.outputs[0].set_type(cls_type([
            N, operator.raw_operator.coef_.shape[0]]))
    else:
        operator.outputs[0].set_type(cls_type([N, 1]))

    # decision_path, decision_leaf
    for n in range(1, len(operator.outputs)):
        operator.outputs[n].type.shape = [N, 1]



# SPDX-License-Identifier: Apache-2.0

from ..common._apply_operation import apply_identity
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from .._supported_operators import sklearn_operator_name_map


def convert_sklearn_grid_search_cv(scope: Scope, operator: Operator,
                                   container: ModelComponentContainer):
    """
    Converter for scikit-learn's GridSearchCV.
    """
    opts = scope.get_options(operator.raw_operator)
    grid_search_op = operator.raw_operator
    best_estimator = grid_search_op.best_estimator_
    op_type = sklearn_operator_name_map[type(best_estimator)]
    grid_search_operator = scope.declare_local_operator(
        op_type, best_estimator)
    container.add_options(id(best_estimator), opts)
    scope.add_options(id(best_estimator), opts)
    grid_search_operator.inputs = operator.inputs

    for i, o in enumerate(operator.outputs):
        v = scope.declare_local_variable(o.onnx_name, type=o.type)
        grid_search_operator.outputs.append(v)
        apply_identity(scope, v.full_name, o.full_name, container)


register_converter('SklearnGridSearchCV',
                   convert_sklearn_grid_search_cv,
                   options="passthrough")



# SPDX-License-Identifier: Apache-2.0

import pprint
from collections import OrderedDict
import hashlib
import numpy as np
from onnx.numpy_helper import from_array
from onnxconverter_common.utils import sklearn_installed, skl2onnx_installed  # noqa
from onnxconverter_common.utils import is_numeric_type, is_string_type  # noqa
from onnxconverter_common.utils import cast_list, convert_to_python_value  # noqa
from onnxconverter_common.utils import convert_to_python_default_value  # noqa
from onnxconverter_common.utils import convert_to_list  # noqa
from onnxconverter_common.utils import check_input_and_output_numbers  # noqa
from onnxconverter_common.utils import check_input_and_output_types  # noqa
from .data_types import TensorType

_unique_index = {'subgraph': 0}


def get_unique_subgraph():
    "Returns a unique identifier integer for subgraph."
    global _unique_index
    _unique_index['subgraph'] += 1
    return _unique_index['subgraph']


def get_producer():
    """
    Internal helper function to return the producer
    """
    from .. import __producer__
    return __producer__


def get_producer_version():
    """
    Internal helper function to return the producer version
    """
    from .. import __producer_version__
    return __producer_version__


def get_domain():
    """
    Internal helper function to return the model domain
    """
    from .. import __domain__
    return __domain__


def get_model_version():
    """
    Internal helper function to return the model version
    """
    from .. import __model_version__
    return __model_version__


def get_column_index(i, inputs):
    """
    Returns a tuples (variable index, column index in that variable).
    The function has two different behaviours, one when *i* (column index)
    is an integer, another one when *i* is a string (column name).
    If *i* is a string, the function looks for input name with
    this name and returns (index, 0).
    If *i* is an integer, let's assume first we have two inputs
    *I0 = FloatTensorType([None, 2])* and *I1 = FloatTensorType([None, 3])*,
    in this case, here are the results:

    ::

        get_column_index(0, inputs) -> (0, 0)
        get_column_index(1, inputs) -> (0, 1)
        get_column_index(2, inputs) -> (1, 0)
        get_column_index(3, inputs) -> (1, 1)
        get_column_index(4, inputs) -> (1, 2)
    """
    if isinstance(i, int):
        if i == 0:
            # Useful shortcut, skips the case when end is None
            # (unknown dimension)
            return 0, 0
        vi = 0
        pos = 0
        end = (inputs[0].type.shape[1]
               if isinstance(inputs[0].type, TensorType) else 1)
        if end is None:
            raise RuntimeError("Cannot extract a specific column {0} when "
                               "one input ('{1}') has unknown "
                               "dimension.".format(i, inputs[0]))
        while True:
            if pos <= i < end:
                return (vi, i - pos)
            vi += 1
            pos = end
            if vi >= len(inputs):
                raise RuntimeError(
                    "Input {} (i={}, end={}) is not available in\n{}".format(
                        vi, i, end, pprint.pformat(inputs)))
            rel_end = (inputs[vi].type.shape[1]
                       if isinstance(inputs[vi].type, TensorType) else 1)
            if rel_end is None:
                raise RuntimeError("Cannot extract a specific column {0} when "
                                   "one input ('{1}') has unknown "
                                   "dimension.".format(i, inputs[vi]))
            end += rel_end
    else:
        for ind, inp in enumerate(inputs):
            if inp.onnx_name == i:
                return ind, 0
        raise RuntimeError(
            "Unable to find column name %r among names %r. "
            "Make sure the input names specified with parameter "
            "initial_types fits the column names specified in the "
            "pipeline to convert. This may happen because a "
            "ColumnTransformer follows a transformer without "
            "any mapped converter in a pipeline." % (
                i, [n.onnx_name for n in inputs]))


def get_column_indices(indices, inputs, multiple):
    """
    Returns the requested graph inpudes based on their
    indices or names. See :func:`get_column_index`.

    :param indices: variables indices or names
    :param inputs: graph inputs
    :param multiple: allows column to come from multiple variables
    :return: a tuple *(variable name, list of requested indices)* if
        *multiple* is False, a dictionary *{ var_index: [ list of
        requested indices ] }*
        if *multiple* is True
    """
    if multiple:
        res = OrderedDict()
        for p in indices:
            ov, onnx_i = get_column_index(p, inputs)
            if ov not in res:
                res[ov] = []
            res[ov].append(onnx_i)
        return res
    else:
        onnx_var = None
        onnx_is = []
        for p in indices:
            ov, onnx_i = get_column_index(p, inputs)
            onnx_is.append(onnx_i)
            if onnx_var is None:
                onnx_var = ov
            elif onnx_var != ov:
                cols = [onnx_var, ov]
                raise NotImplementedError(
                    "sklearn-onnx is not able to merge multiple columns from "
                    "multiple variables ({0}). You should think about merging "
                    "initial types.".format(cols))
        return onnx_var, onnx_is


def hash_array(value, length=15):
    "Computes a hash identifying the value."
    try:
        onx = from_array(value)
    except AttributeError as e:
        # sparse matrix for example
        if hasattr(value, 'tocoo'):
            coo = value.tocoo()
            arrs = [coo.data, coo.row, coo.col, np.array(coo.shape)]
            m = hashlib.sha256()
            for arr in arrs:
                m.update(from_array(arr).SerializeToString())
            return m.hexdigest()[:length]

        raise ValueError(
            "Unable to compute hash for type %r (value=%r)." % (
                type(value), value)) from e
    except RuntimeError as ee:
        # cannot be serialized
        if isinstance(value, (np.ndarray, list)):
            b = str(value).encode('utf-8')
            m = hashlib.sha256()
            m.update(b)
            return m.hexdigest()[:length]
        raise RuntimeError(
            "Unable to convert value type %r, (value=%r)." % (
                type(value), value)) from ee

    m = hashlib.sha256()
    m.update(onx.SerializeToString())
    return m.hexdigest()[:length]



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_shape_calculator
from ..common.data_types import Int64TensorType


def calculate_isolation_forest_output_shapes(operator):
    N = operator.inputs[0].get_first_dimension()
    operator.outputs[0].type = Int64TensorType([N, 1])
    operator.outputs[1].type.shape = [N, 1]


register_shape_calculator(
    'SklearnIsolationForest', calculate_isolation_forest_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from sklearn.preprocessing import normalize
from ..algebra.complex_functions import onnx_cdist
from ..algebra.onnx_ops import (
    OnnxMatMul, OnnxTranspose, OnnxDiv, OnnxSub, OnnxAdd,
    OnnxMul, OnnxPow, OnnxTanh, OnnxSqrt, OnnxExp,
    OnnxReduceSumApi11)
from ..algebra.onnx_operator import OnnxSubEstimator
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_numpy_type


def kernel_centerer_converter(scope: Scope, operator: Operator,
                              container: ModelComponentContainer):
    op = operator.raw_operator
    op_version = container.target_opset
    X = operator.inputs[0]
    dtype = guess_numpy_type(X.type)

    N = np.array([op.K_fit_rows_.shape[0]], dtype=dtype)
    K_pred_cols = OnnxDiv(
            OnnxReduceSumApi11(X, axes=[1], op_version=op_version),
            N, op_version=op_version)

    # K -= self.K_fit_rows_
    # K -= K_pred_cols
    # K += self.K_fit_all_
    K1 = OnnxSub(X, op.K_fit_rows_.astype(dtype), op_version=op_version)
    K2 = OnnxSub(K1, K_pred_cols, op_version=op_version)
    final = OnnxAdd(K2, np.array([op.K_fit_all_], dtype=dtype),
                    op_version=op_version,
                    output_names=operator.outputs[:1])
    final.add_to(scope, container)


def kernel_pca_converter(scope: Scope, operator: Operator,
                         container: ModelComponentContainer):
    op = operator.raw_operator
    op_version = container.target_opset
    X = operator.inputs[0]
    dtype = guess_numpy_type(X.type)
    options = container.get_options(op, dict(optim=None))
    optim = options['optim']

    # def _get_kernel(self, X, Y=None):
    # return pairwise_kernels(
    #         X, Y, metric=self.kernel, filter_params=True, **params)
    if callable(op.kernel):
        raise RuntimeError(
            "Unable to converter KernelPCA with a custom kernel %r."
            "" % op.kernel)
    if op.kernel == 'precomputed':
        raise RuntimeError(
            "The converter is not implemented when kernel=%r for "
            "type=%r." % (op.kernel, type(op)))

    kernel = op.kernel
    params = {"gamma": op.gamma, "degree": op.degree, "coef0": op.coef0}

    if kernel == 'linear':
        Y = op.X_fit_.astype(dtype)
        dist = OnnxMatMul(
            X, OnnxTranspose(Y, perm=[1, 0], op_version=op_version),
            op_version=op_version)
    elif kernel == 'cosine':
        yn = normalize(op.X_fit_, copy=True)
        ynt = yn.astype(dtype)
        norm = OnnxSqrt(
            OnnxReduceSumApi11(
                OnnxPow(X, np.array([2], dtype=np.int64),
                        op_version=op_version),
                axes=[1], op_version=op_version, keepdims=1),
            op_version=op_version)
        dist = OnnxMatMul(
            OnnxDiv(X, norm, op_version=op_version),
            OnnxTranspose(ynt, perm=[1, 0], op_version=op_version),
            op_version=op_version)
    elif kernel in ('poly', 'sigmoid'):
        Y = op.X_fit_.astype(dtype)
        dot = OnnxMatMul(
            X, OnnxTranspose(Y, perm=[1, 0], op_version=op_version),
            op_version=op_version)
        if params['gamma'] is None:
            gamma = np.array([1. / Y.shape[1]], dtype=dtype)
        else:
            gamma = np.array([params['gamma']], dtype=dtype)
        dot_g = OnnxMul(dot, gamma, op_version=op_version)
        dot_c = OnnxAdd(dot_g, np.array([params['coef0']], dtype=dtype),
                        op_version=op_version)
        if kernel == 'poly':
            dist = OnnxPow(dot_c,
                           np.array([params['degree']], dtype=np.int64),
                           op_version=op_version)
        else:
            dist = OnnxTanh(dot_c, op_version=op_version)
    elif kernel == 'rbf':
        if optim == 'cdist':
            from skl2onnx.algebra.custom_ops import OnnxCDist
            Y = op.X_fit_.astype(dtype)
            pair = OnnxCDist(X, Y, metric='sqeuclidean', op_version=op_version)
        elif optim is None:
            Y = op.X_fit_.astype(dtype)
            dim_in = Y.shape[1] if hasattr(Y, 'shape') else None
            dim_out = Y.shape[0] if hasattr(Y, 'shape') else None
            pair = onnx_cdist(X, Y, metric='sqeuclidean', dtype=dtype,
                              op_version=op_version,
                              dim_in=dim_in, dim_out=dim_out)
        else:
            raise ValueError("Unknown optimisation '{}'.".format(optim))
        if params['gamma'] is None:
            gamma = np.array([-1. / Y.shape[1]], dtype=dtype)
        else:
            gamma = np.array([-params['gamma']], dtype=dtype)
        pair_g = OnnxMul(pair, gamma, op_version=op_version)
        dist = OnnxExp(pair_g, op_version=op_version)
    else:
        raise ValueError("Unknown kernel '{}'.".format(kernel))

    #  K = self._centerer.transform(self._get_kernel(X, self.X_fit_))
    K = OnnxSubEstimator(op._centerer, dist, op_version=op_version)

    if hasattr(op, 'eigenvalues_'):
        # scikit-learn>=1.0
        non_zeros = np.flatnonzero(op.eigenvalues_)
        scaled_alphas = np.zeros_like(op.eigenvectors_)
        scaled_alphas[:, non_zeros] = (
            op.eigenvectors_[:, non_zeros] /
            np.sqrt(op.eigenvalues_[non_zeros]))
    else:
        # scikit-learn<1.0
        non_zeros = np.flatnonzero(op.lambdas_)
        scaled_alphas = np.zeros_like(op.alphas_)
        scaled_alphas[:, non_zeros] = (
            op.alphas_[:, non_zeros] / np.sqrt(op.lambdas_[non_zeros]))

    # np.dot(K, scaled_alphas)
    output = OnnxMatMul(K, scaled_alphas.astype(dtype),
                        op_version=op_version,
                        output_names=operator.outputs[:1])

    # register the output
    output.add_to(scope, container)


register_converter('SklearnKernelCenterer', kernel_centerer_converter)
register_converter('SklearnKernelPCA', kernel_pca_converter,
                   options={'optim': [None, 'cdist']})



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from sklearn.base import TransformerMixin, BaseEstimator
try:
    from sklearn.utils.validation import _deprecate_positional_args
except ImportError:
    def _deprecate_positional_args(x): return x  # noqa


class CastTransformer(TransformerMixin, BaseEstimator):

    """
    Cast features into a specific types.
    This should be used to minimize the conversion
    of a pipeline using float32 instead of double.

    Parameters
    ----------
    dtype : numpy type,
        output are cast into that type
    """  # noqa

    @_deprecate_positional_args
    def __init__(self, *, dtype=np.float32):
        self.dtype = dtype

    def _cast(self, a, name):
        if not isinstance(a, np.ndarray):
            if hasattr(a, 'values') and hasattr(a, 'iloc'):
                # dataframe
                a = a.values
            elif not hasattr(a, 'astype'):
                raise TypeError(
                    "{} must be a numpy array or a dataframe.".format(
                        name))
        try:
            a2 = a.astype(self.dtype)
        except ValueError:
            raise ValueError(
                "Unable to cast {} from {} into {}.".format(
                    name, a.dtype, self.dtype))
        return a2

    def fit(self, X, y=None, sample_weight=None):
        """
        Does nothing except checking *dtype* may be applied.
        """
        self._cast(X, 'X')
        return self

    def transform(self, X, y=None):
        """
        Casts array X.
        """
        return self._cast(X, 'X')



# SPDX-License-Identifier: Apache-2.0


import copy
from ..common._registration import register_shape_calculator
from ..common.data_types import FloatTensorType
from ..common.data_types import Int64TensorType, StringTensorType
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_label_encoder_output_shapes(operator):
    """
    This function just copy the input shape to the output because label
    encoder only alters input features' values, not their shape.
    """
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[
                                 FloatTensorType, Int64TensorType,
                                 StringTensorType])

    input_shape = copy.deepcopy(operator.inputs[0].type.shape)
    operator.outputs[0].type = Int64TensorType(copy.deepcopy(input_shape))


register_shape_calculator('SklearnLabelEncoder',
                          calculate_sklearn_label_encoder_output_shapes)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_converter
from ..common._apply_operation import apply_concat, apply_identity
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_function_transformer(scope: Scope, operator: Operator,
                                         container: ModelComponentContainer):
    op = operator.raw_operator
    if op.func is not None:
        raise RuntimeError("FunctionTransformer is not supported unless the "
                           "transform function is None (= identity). "
                           "You may raise an issue at "
                           "https://github.com/onnx/sklearn-onnx/issues.")
    if len(operator.inputs) == 1:
        apply_identity(scope, operator.inputs[0].full_name,
                       operator.outputs[0].full_name, container)
    else:
        apply_concat(scope, [i.full_name for i in operator.inputs],
                     operator.outputs[0].full_name, container)


register_converter('SklearnFunctionTransformer',
                   convert_sklearn_function_transformer)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import (
    FloatTensorType
)


def powertransformer_shape_calculator(operator):
    """Shape calculator for PowerTransformer"""
    inputs = operator.inputs[0]
    output = operator.outputs[0]
    n, c = inputs.type.shape
    output.type = FloatTensorType([n, c])


register_shape_calculator('SklearnPowerTransformer',
                          powertransformer_shape_calculator)



# SPDX-License-Identifier: Apache-2.0


import numbers
import numpy as np
from ..common._apply_operation import (
    apply_cast,
    apply_concat,
    apply_div,
    apply_mul,
    apply_reshape,
    apply_reducesum,
    apply_transpose,
)
from ..common._registration import register_converter
from ..common.data_types import (
    BooleanTensorType, Int64TensorType, guess_numpy_type,
    guess_proto_type)
from ..common.tree_ensemble import (
    add_tree_to_attribute_pairs,
    get_default_tree_classifier_attribute_pairs,
    get_default_tree_regressor_attribute_pairs,
)
from ..common.utils_classifier import get_label_classes
from ..proto import onnx_proto


def populate_tree_attributes(model, name, dtype):
    """Construct attrs dictionary to be used in predict()
    while adding a node with TreeEnsembleClassifier ONNX op.
    """
    attrs = {}
    attrs['name'] = name
    attrs['post_transform'] = 'NONE'
    attrs['nodes_treeids'] = []
    attrs['nodes_nodeids'] = []
    attrs['nodes_featureids'] = []
    attrs['nodes_modes'] = []
    attrs['nodes_values'] = []
    attrs['nodes_truenodeids'] = []
    attrs['nodes_falsenodeids'] = []
    attrs['nodes_missing_value_tracks_true'] = []
    attrs['nodes_hitrates'] = []
    attrs['class_treeids'] = []
    attrs['class_nodeids'] = []
    attrs['class_ids'] = []
    attrs['class_weights'] = []
    attrs['classlabels_int64s'] = list(range(model.tree_.node_count))

    for i in range(model.tree_.node_count):
        node_id = i
        if (model.tree_.children_left[i] > i and
                model.tree_.children_right[i] > i):
            feat = model.tree_.feature[i]
            thresh = model.tree_.threshold[i]
            left = model.tree_.children_left[i]
            right = model.tree_.children_right[i]
            mode = 'BRANCH_LEQ'
        else:
            feat, thresh, left, right = 0, 0., 0, 0
            mode = 'LEAF'
        attrs['nodes_nodeids'].append(node_id)
        attrs['nodes_treeids'].append(0)
        attrs['nodes_featureids'].append(feat)
        attrs['nodes_modes'].append(mode)
        attrs['nodes_truenodeids'].append(left)
        attrs['nodes_falsenodeids'].append(right)
        attrs['nodes_missing_value_tracks_true'].append(False)
        attrs['nodes_hitrates'].append(1.)
        attrs['nodes_values'].append(thresh)
        if mode == 'LEAF':
            attrs['class_ids'].append(node_id)
            attrs['class_weights'].append(1.)
            attrs['class_treeids'].append(0)
            attrs['class_nodeids'].append(node_id)
    if dtype is not None:
        for k in attrs:
            if k in ('node_values', 'class_weights', 'target_weights'):
                attrs[k] = np.array(attrs[k], dtype=dtype)
    return attrs


def predict(model, scope, operator, container,
            op_type, op_domain, op_version, is_ensemble=False):
    """Predict target and calculate probability scores."""
    indices_name = scope.get_unique_variable_name('indices')
    dummy_proba_name = scope.get_unique_variable_name('dummy_proba')
    values_name = scope.get_unique_variable_name('values')
    out_values_name = scope.get_unique_variable_name('out_indices')
    transposed_result_name = scope.get_unique_variable_name(
        'transposed_result')
    proba_output_name = scope.get_unique_variable_name('proba_output')
    cast_result_name = scope.get_unique_variable_name('cast_result')
    reshaped_indices_name = scope.get_unique_variable_name('reshaped_indices')
    sum_output_name = scope.get_unique_variable_name('sum_proba')
    value = model.tree_.value.transpose(1, 2, 0)

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32

    container.add_initializer(
        values_name, proto_dtype, value.shape, value.ravel())

    input_name = operator.input_full_names
    if type(operator.inputs[0].type) == BooleanTensorType:
        cast_input_name = scope.get_unique_variable_name('cast_input')

        apply_cast(scope, input_name, cast_input_name,
                   container, to=proto_dtype)
        input_name = cast_input_name

    if model.tree_.node_count > 1:
        attrs = populate_tree_attributes(
            model, scope.get_unique_operator_name(op_type), dtype)
        container.add_node(
            op_type, input_name,
            [indices_name, dummy_proba_name],
            op_domain=op_domain, op_version=op_version, **attrs)
    else:
        zero_name = scope.get_unique_variable_name('zero')
        zero_matrix_name = scope.get_unique_variable_name('zero_matrix')
        reduced_zero_matrix_name = scope.get_unique_variable_name(
            'reduced_zero_matrix')

        container.add_initializer(
            zero_name, proto_dtype, [], [0])
        apply_mul(scope, [input_name[0], zero_name],
                  zero_matrix_name, container, broadcast=1)
        if container.target_opset < 13:
            container.add_node(
                'ReduceSum', zero_matrix_name, reduced_zero_matrix_name,
                axes=[1], name=scope.get_unique_operator_name('ReduceSum'))
        else:
            axis_name = scope.get_unique_variable_name('axis')
            container.add_initializer(
                axis_name, onnx_proto.TensorProto.INT64, [1], [1])
            container.add_node(
                'ReduceSum', [zero_matrix_name, axis_name],
                reduced_zero_matrix_name,
                name=scope.get_unique_operator_name('ReduceSum'))
        apply_cast(scope, reduced_zero_matrix_name, indices_name,
                   container, to=onnx_proto.TensorProto.INT64)
    apply_reshape(scope, indices_name, reshaped_indices_name,
                  container, desired_shape=[1, -1])
    container.add_node(
        'ArrayFeatureExtractor',
        [values_name, reshaped_indices_name],
        out_values_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
    apply_transpose(scope, out_values_name, proba_output_name,
                    container, perm=(0, 2, 1))

    if is_ensemble:
        proba_result_name = scope.get_unique_variable_name('proba_result')
        apply_reducesum(scope, proba_output_name, sum_output_name,
                        container, keepdims=1, axes=[2])
        apply_div(scope, [proba_output_name, sum_output_name],
                  proba_result_name, container)
        return proba_result_name
    else:
        apply_cast(scope, proba_output_name, cast_result_name,
                   container, to=onnx_proto.TensorProto.BOOL)
        apply_cast(scope, cast_result_name, operator.outputs[1].full_name,
                   container, to=proto_dtype)
        apply_transpose(scope, out_values_name, transposed_result_name,
                        container, perm=(2, 1, 0))
        return transposed_result_name


def _append_decision_output(
        input_name, attrs, fct_label, n_out, scope, operator, container,
        op_type='TreeEnsembleClassifier',
        op_domain='ai.onnx.ml', op_version=1,
        cast_encode=False, regression=False, dtype=np.float32,
        overwrite_tree=None):

    attrs = attrs.copy()
    attrs['name'] = scope.get_unique_operator_name(op_type)
    attrs['n_targets'] = 1
    attrs['post_transform'] = 'NONE'
    if regression:
        attrs['target_weights'] = np.array(
            [float(_) for _ in attrs['target_nodeids']], dtype=dtype)
    else:
        attrs['target_ids'] = [0 for _ in attrs['class_ids']]
        attrs['target_weights'] = [float(_) for _ in attrs['class_nodeids']]
        attrs['target_nodeids'] = attrs['class_nodeids']
        attrs['target_treeids'] = attrs['class_treeids']

    rem = [k for k in attrs if k.startswith('class')]
    for k in rem:
        del attrs[k]
    dpath = scope.get_unique_variable_name("dpath")
    container.add_node(
        op_type.replace("Classifier", "Regressor"), input_name, dpath,
        op_domain=op_domain, op_version=op_version, **attrs)

    if n_out is None:
        final_name = scope.get_unique_variable_name("dpatho")
    else:
        final_name = operator.outputs[n_out].full_name

    if cast_encode:
        apply_cast(
            scope, dpath, final_name,
            container, to=onnx_proto.TensorProto.INT64,
            operator_name=scope.get_unique_operator_name('TreePathType'))
    else:
        op = operator.raw_operator
        labels = fct_label(
            overwrite_tree if overwrite_tree is not None else op.tree_)
        ordered = list(sorted(labels.items()))
        keys = [float(_[0]) for _ in ordered]
        values = [_[1] for _ in ordered]
        name = scope.get_unique_variable_name("spath")
        container.add_node(
            'LabelEncoder', dpath, name,
            op_domain=op_domain, op_version=2,
            default_string='0', keys_floats=keys, values_strings=values,
            name=scope.get_unique_operator_name('TreePath'))
        apply_reshape(
            scope, name, final_name,
            container, desired_shape=(-1, 1),
            operator_name=scope.get_unique_operator_name('TreePathShape'))

    return final_name


def convert_sklearn_decision_tree_classifier(
        scope, operator, container, op_type='TreeEnsembleClassifier',
        op_domain='ai.onnx.ml', op_version=1):
    try:
        dtype = guess_numpy_type(operator.inputs[0].type)
    except NotImplementedError as e:
        raise RuntimeError(
            "Unknown variable {}.".format(operator.inputs[0])) from e
    if dtype != np.float64:
        dtype = np.float32
    op = operator.raw_operator
    options = scope.get_options(
        op, dict(decision_path=False, decision_leaf=False))
    if op.n_outputs_ == 1:
        attrs = get_default_tree_classifier_attribute_pairs()
        attrs['name'] = scope.get_unique_operator_name(op_type)
        classes = get_label_classes(scope, op)

        if all(isinstance(i, np.ndarray) for i in classes):
            classes = np.concatenate(classes)
        if all(isinstance(i, (numbers.Real, bool, np.bool_)) for i in classes):
            class_labels = [int(i) for i in classes]
            attrs['classlabels_int64s'] = class_labels
        elif all(isinstance(i, str) for i in classes):
            class_labels = [str(i) for i in classes]
            attrs['classlabels_strings'] = class_labels
        else:
            raise ValueError('Labels must be all integers or all strings.')

        add_tree_to_attribute_pairs(attrs, True, op.tree_, 0, 1., 0, True,
                                    True, dtype=dtype)
        input_name = operator.input_full_names
        if type(operator.inputs[0].type) == BooleanTensorType:
            cast_input_name = scope.get_unique_variable_name('cast_input')

            apply_cast(scope, input_name, cast_input_name,
                       container, to=onnx_proto.TensorProto.FLOAT)
            input_name = cast_input_name

        if dtype is not None:
            for k in attrs:
                if k in ('nodes_values', 'class_weights',
                         'target_weights', 'nodes_hitrates',
                         'base_values'):
                    attrs[k] = np.array(attrs[k], dtype=dtype)

        container.add_node(
            op_type, input_name,
            [operator.outputs[0].full_name, operator.outputs[1].full_name],
            op_domain=op_domain, op_version=op_version, **attrs)

        n_out = 2
        if options['decision_path']:
            # decision_path
            _append_decision_output(
                input_name, attrs, _build_labels_path, n_out,
                scope, operator, container,
                op_type=op_type, op_domain=op_domain,
                op_version=op_version, dtype=dtype)
            n_out += 1
        if options['decision_leaf']:
            # decision_path
            _append_decision_output(
                input_name, attrs, _build_labels_leaf, n_out,
                scope, operator, container,
                op_type=op_type, op_domain=op_domain,
                op_version=op_version, cast_encode=True,
                dtype=dtype)
            n_out += 1
    else:
        transposed_result_name = predict(
            op, scope, operator, container, op_type, op_domain, op_version)
        predictions = []
        for k in range(op.n_outputs_):
            preds_name = scope.get_unique_variable_name('preds')
            reshaped_preds_name = scope.get_unique_variable_name(
                'reshaped_preds')
            k_name = scope.get_unique_variable_name('k_column')
            out_k_name = scope.get_unique_variable_name('out_k_column')
            argmax_output_name = scope.get_unique_variable_name(
                'argmax_output')
            classes_name = scope.get_unique_variable_name('classes')
            reshaped_result_name = scope.get_unique_variable_name(
                'reshaped_result')

            container.add_initializer(
                k_name, onnx_proto.TensorProto.INT64,
                [], [k])
            container.add_initializer(
                classes_name, onnx_proto.TensorProto.INT64,
                op.classes_[k].shape, op.classes_[k])

            container.add_node(
                'ArrayFeatureExtractor', [transposed_result_name, k_name],
                out_k_name, op_domain='ai.onnx.ml',
                name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
            container.add_node(
                'ArgMax', out_k_name, argmax_output_name,
                name=scope.get_unique_operator_name('ArgMax'), axis=1)
            apply_reshape(scope, argmax_output_name, reshaped_result_name,
                          container, desired_shape=(1, -1))
            container.add_node(
                'ArrayFeatureExtractor', [classes_name, reshaped_result_name],
                preds_name, op_domain='ai.onnx.ml',
                name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
            apply_reshape(scope, preds_name, reshaped_preds_name,
                          container, desired_shape=(-1, 1))
            predictions.append(reshaped_preds_name)
        apply_concat(scope, predictions, operator.outputs[0].full_name,
                     container, axis=1)

        if options['decision_path']:
            raise RuntimeError(
                "Option decision_path for multi-outputs "
                "is not implemented yet.")
        if options['decision_leaf']:
            raise RuntimeError(
                "Option decision_leaf for multi-outputs "
                "is not implemented yet.")


def convert_sklearn_decision_tree_regressor(
        scope, operator, container, op_type='TreeEnsembleRegressor',
        op_domain='ai.onnx.ml', op_version=1):
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    op = operator.raw_operator

    attrs = get_default_tree_regressor_attribute_pairs()
    attrs['name'] = scope.get_unique_operator_name(op_type)
    attrs['n_targets'] = int(op.n_outputs_)
    add_tree_to_attribute_pairs(attrs, False, op.tree_, 0, 1., 0, False,
                                True, dtype=dtype)

    if dtype is not None:
        for k in attrs:
            if k in ('nodes_values', 'class_weights',
                     'target_weights', 'nodes_hitrates',
                     'base_values'):
                attrs[k] = np.array(attrs[k], dtype=dtype)

    input_name = operator.input_full_names
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name('cast_input')

        apply_cast(scope, operator.input_full_names, cast_input_name,
                   container, to=onnx_proto.TensorProto.FLOAT)
        input_name = [cast_input_name]

    container.add_node(
        op_type, input_name, operator.outputs[0].full_name,
        op_domain=op_domain, op_version=op_version, **attrs)

    options = scope.get_options(
        op, dict(decision_path=False, decision_leaf=False))

    # decision_path
    n_out = 1
    if options['decision_path']:
        # decision_path
        _append_decision_output(
            input_name, attrs, _build_labels_path, n_out,
            scope, operator, container,
            op_type=op_type, op_domain=op_domain,
            op_version=op_version, regression=True)
        n_out += 1
    if options['decision_leaf']:
        # decision_path
        _append_decision_output(
            input_name, attrs, _build_labels_leaf, n_out,
            scope, operator, container,
            op_type=op_type, op_domain=op_domain,
            op_version=op_version, regression=True, cast_encode=True)
        n_out += 1


def _recursive_build_labels(tree, index, current):
    current[index] = True
    if tree.children_left[index] == -1:
        yield (index, current.copy())
    else:
        for it in _recursive_build_labels(
                tree, tree.children_left[index], current):
            yield it
        for it in _recursive_build_labels(
                tree, tree.children_right[index], current):
            yield it
    current[index] = False


def _build_labels_path(tree):
    paths = {}
    current = {}

    for leave_index, path in _recursive_build_labels(tree, 0, current):
        spath = ["0" for _ in range(tree.node_count)]
        for nodeid, b in path.items():
            if b:
                spath[nodeid] = "1"
        paths[leave_index] = ''.join(spath)
    return paths


def _build_labels_leaf(tree):
    paths = {}
    current = {}

    for leave_index, path in _recursive_build_labels(tree, 0, current):
        paths[leave_index] = leave_index
    return paths


register_converter('SklearnDecisionTreeClassifier',
                   convert_sklearn_decision_tree_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'decision_path': [True, False],
                            'decision_leaf': [True, False]})
register_converter('SklearnDecisionTreeRegressor',
                   convert_sklearn_decision_tree_regressor,
                   options={'decision_path': [True, False],
                            'decision_leaf': [True, False]})
register_converter('SklearnExtraTreeClassifier',
                   convert_sklearn_decision_tree_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'decision_path': [True, False],
                            'decision_leaf': [True, False]})
register_converter('SklearnExtraTreeRegressor',
                   convert_sklearn_decision_tree_regressor,
                   options={'decision_path': [True, False],
                            'decision_leaf': [True, False]})



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from sklearn.utils.extmath import row_norms
from ..common.data_types import Int64TensorType, guess_numpy_type
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..algebra.onnx_ops import (
    OnnxReduceSumSquare, OnnxGemm, OnnxMatMul,
    OnnxAdd, OnnxArgMin, OnnxCast, OnnxSqrt, OnnxMul)


def convert_sklearn_kmeans(scope: Scope, operator: Operator,
                           container: ModelComponentContainer):
    """
    Computation graph of distances to all centroids for a batch of examples.
    Note that a centriod is just the center of a cluster. We use ``[]`` to
    denote the dimension of a variable; for example, ``X[3, 2]`` means that
    *X* is a *3-by-2* tensor. In addition, for a matrix *X*, $X'$ denotes its
    transpose.

    Symbols:

    * *l*: # of examples.
    * *n*: # of features per input example.
    * *X*: input examples, l-by-n tensor.
    * *C*: centroids, k-by-n tensor.
    * :math:`C^2`: 2-norm of all centriod vectors, its shape is ``[k]``.
    * *Y*: 2-norm of difference between examples and centroids,
      *l-by-k* tensor. The value at i-th row and k-th column row,
      ``Y[i,k]``,is the distance from example *i* to centroid *k*.
    * *L*: the id of the nearest centroid for each input example,
      its shape is ``[l]``.

    ::

         .------------------------------------------------------.
         |                                                      |
         |                                                      v
        X [l, n] --> ReduceSumSquare -> X^2 [l]   Gemm (alpha=-2, transB=1)
                                         |                  |  |- C [k, n]
                                         |                  |
                                         |                  v
                                         `------> Add <-- -2XC' [l, k]
                                                   |
                                                   v
                     C^2 [k] --------> Add <----- Z [l, k]
                                        |
                                        v
                 L [l] <-- ArgMin <--  Y2 [l, k] --> Sqrt --> Y2 [l, k]

    *scikit-learn* code:

    ::

        X = data
        Y = model.cluster_centers_
        XX = row_norms(X, squared=True)
        YY = row_norms(Y, squared=True)
        distances = safe_sparse_dot(X, Y.T, dense_output=True)
        distances *= -2
        distances += XX[:, numpy.newaxis]
        distances += YY[numpy.newaxis, :]
        numpy.sqrt(distances, out=distances)
    """
    X = operator.inputs[0]
    out = operator.outputs
    op = operator.raw_operator
    options = container.get_options(op, dict(gemm=True))
    opv = container.target_opset
    C = op.cluster_centers_
    input_name = X
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32

    if type(X.type) == Int64TensorType:
        x_cast = OnnxCast(X, to=np.float32, op_version=opv)
        input_name = x_cast

    C2 = row_norms(C, squared=True).astype(dtype)
    C = C.astype(dtype)
    rs = OnnxReduceSumSquare(input_name, axes=[1], keepdims=1, op_version=opv)

    if options['gemm']:
        N = X.get_first_dimension()
        if isinstance(N, int):
            zeros = np.zeros((N, ), dtype=dtype)
        else:
            zeros = OnnxMul(rs, np.array([0], dtype=dtype),
                            op_version=opv)
        gemm_out = OnnxGemm(input_name, C, zeros, alpha=-2.,
                            transB=1, op_version=opv)
    else:
        gemm_out = OnnxMatMul(
            input_name, (C.T * (-2)).astype(dtype), op_version=opv)

    z = OnnxAdd(rs, gemm_out, op_version=opv)
    y2 = OnnxAdd(C2, z, op_version=opv)
    ll = OnnxArgMin(y2, axis=1, keepdims=0, output_names=out[:1],
                    op_version=opv)
    y2s = OnnxSqrt(y2, output_names=out[1:], op_version=opv)
    ll.add_to(scope, container)
    y2s.add_to(scope, container)


register_converter('SklearnKMeans', convert_sklearn_kmeans,
                   options={'gemm': [True, False]})
register_converter('SklearnMiniBatchKMeans', convert_sklearn_kmeans,
                   options={'gemm': [True, False]})



# SPDX-License-Identifier: Apache-2.0


import numpy as np

from ..proto import onnx_proto
from ..common._apply_operation import (
    apply_cast, apply_concat, apply_reshape,
    apply_mul, apply_add
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_k_bins_discretiser(scope: Scope, operator: Operator,
                                       container: ModelComponentContainer):
    op = operator.raw_operator

    if op.encode == 'onehot':
        raise RuntimeError("onehot encoding not supported. "
                           "ONNX does not support sparse tensors. "
                           "with opset < 11. You may raise an isue at "
                           "https://github.com/onnx/sklearn-onnx/issues.")

    ranges = list(map(lambda e: e[1:-1] if len(e) > 2
                      else [np.finfo(np.float32).max], op.bin_edges_))
    digitised_output_name = [None] * len(ranges)
    last_column_name = None

    for i, item in enumerate(ranges):
        digitised_output_name[i] = (
            scope.get_unique_variable_name('digitised_output_{}'.format(i)))
        column_index_name = scope.get_unique_variable_name('column_index')
        range_column_name = scope.get_unique_variable_name('range_column')
        column_name = scope.get_unique_variable_name('column')
        cast_column_name = scope.get_unique_variable_name('cast_column')
        less_result_name = scope.get_unique_variable_name('less_result')
        cast_result_name = scope.get_unique_variable_name('cast_result')
        concatenated_array_name = scope.get_unique_variable_name(
            'concatenated_array')
        argmax_output_name = scope.get_unique_variable_name('argmax_output')

        container.add_initializer(column_index_name,
                                  onnx_proto.TensorProto.INT64, [], [i])
        container.add_initializer(range_column_name,
                                  onnx_proto.TensorProto.FLOAT,
                                  [len(item)], item)

        container.add_node(
            'ArrayFeatureExtractor',
            [operator.inputs[0].full_name, column_index_name], column_name,
            name=scope.get_unique_operator_name('ArrayFeatureExtractor'),
            op_domain='ai.onnx.ml')
        apply_cast(scope, column_name, cast_column_name,
                   container, to=onnx_proto.TensorProto.FLOAT)
        container.add_node(
            'Less', [cast_column_name, range_column_name],
            less_result_name,
            name=scope.get_unique_operator_name('Less'))
        apply_cast(scope, less_result_name, cast_result_name,
                   container, to=onnx_proto.TensorProto.FLOAT)

        if last_column_name is None:
            last_column_name = scope.get_unique_variable_name('last_column')
            zero_float = scope.get_unique_variable_name('zero_float')
            one_float = scope.get_unique_variable_name('one_float')
            zero_column = scope.get_unique_variable_name('zero_column')
            container.add_initializer(
                one_float, onnx_proto.TensorProto.FLOAT,
                [1], np.ones(1))
            container.add_initializer(
                zero_float, onnx_proto.TensorProto.FLOAT,
                [1], np.zeros(1))
            apply_mul(scope, [cast_column_name, zero_float], zero_column,
                      container, broadcast=1)
            apply_add(scope, [zero_column, one_float], last_column_name,
                      container, broadcast=1)

        apply_concat(scope, [cast_result_name, last_column_name],
                     concatenated_array_name, container, axis=1)
        container.add_node('ArgMax', concatenated_array_name,
                           argmax_output_name, axis=1,
                           name=scope.get_unique_operator_name('ArgMax'))
        if op.encode == 'onehot-dense':
            onehot_result_name = scope.get_unique_variable_name(
                'onehot_result')

            container.add_node(
                'OneHotEncoder', argmax_output_name,
                onehot_result_name,
                name=scope.get_unique_operator_name('OneHotEncoder'),
                cats_int64s=list(range(op.n_bins_[i])),
                op_domain='ai.onnx.ml')
            apply_reshape(scope, onehot_result_name, digitised_output_name[i],
                          container, desired_shape=(-1, op.n_bins_[i]))
        else:
            apply_cast(scope, argmax_output_name, digitised_output_name[i],
                       container, to=onnx_proto.TensorProto.FLOAT)
    apply_concat(scope, digitised_output_name,
                 operator.outputs[0].full_name, container, axis=1)


register_converter('SklearnKBinsDiscretizer',
                   convert_sklearn_k_bins_discretiser)



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_shape_calculator


def calculate_sklearn_sequence_at(operator):
    pass


def calculate_sklearn_sequence_construct(operator):
    pass


register_shape_calculator('SklearnSequenceAt', calculate_sklearn_sequence_at)
register_shape_calculator(
    'SklearnSequenceConstruct', calculate_sklearn_sequence_construct)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers


def calculate_sklearn_class_labels(operator):
    check_input_and_output_numbers(operator, output_count_range=1)


register_shape_calculator(
    'SklearnClassLabels', calculate_sklearn_class_labels)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE
from onnxconverter_common.onnx_ops import *  # noqa
from ..proto import onnx_proto


def apply_normalizer(scope, inputs, outputs, container,
                     norm, use_float):
    """
    Adds operator Normalizer if *use_float* is true,
    otherwise, uses *ReduceSum* + *Div*. *Normalizer*
    always produces float according to ONNX speciciations.
    """
    input = inputs[0] if isinstance(inputs, list) else inputs
    output = outputs[0] if isinstance(outputs, list) else outputs
    use_normalizer = container.is_allowed({'Normalizer'})

    if use_normalizer and use_float:
        container.add_node(
            'Normalizer', input, output,
            op_domain='ai.onnx.ml', norm=norm,
            name=scope.get_unique_operator_name('Normalizer'))
    else:
        # Normalizer only produces floats.
        if norm == 'L1':
            norm = scope.get_unique_variable_name('norm')
            norm_abs = scope.get_unique_variable_name('norm_abs')
            container.add_node(
                'Abs', input, norm_abs,
                name=scope.get_unique_operator_name('Abs'))

            if container.target_opset < 13:
                container.add_node(
                    'ReduceSum', norm_abs, norm, axes=[1], keepdims=1,
                    name=scope.get_unique_operator_name('ReduceSum'))
            else:
                axis_name = scope.get_unique_variable_name('axis')
                container.add_initializer(
                    axis_name, onnx_proto.TensorProto.INT64, [1], [1])
                container.add_node(
                    'ReduceSum', [norm_abs, axis_name], norm, keepdims=1,
                    name=scope.get_unique_operator_name('ReduceSum'))
            apply_div(  # noqa
                scope, [input, norm], output, container,
                operator_name=scope.get_unique_operator_name(
                    'NormalizerNorm'))
        elif norm == 'L2':
            norm = scope.get_unique_variable_name('norm')
            norm2 = scope.get_unique_variable_name('norm2')
            container.add_node(
                'ReduceSumSquare', input, norm, axes=[1], keepdims=1,
                name=scope.get_unique_operator_name('ReduceSumSquare'))
            container.add_node(
                'Sqrt', [norm], norm2,
                name=scope.get_unique_operator_name('Sqrt'))
            apply_div(  # noqa
                scope, [input, norm2], output, container,
                operator_name=scope.get_unique_operator_name(
                    'NormalizerNorm'))
        else:
            raise NotImplementedError(
                "Normalization not implemented for norm %r." % norm)


def _create_name_or_use_existing_one(scope, op_type, name):
    if name is None:
        return scope.get_unique_operator_name(op_type)
    return name


def apply_clip(scope, input_name, output_name, container,
               operator_name=None, max=None, min=None):
    name = _create_name_or_use_existing_one(scope, 'Clip', operator_name)
    attrs = {'name': name}

    if container.target_opset < 11:
        if max is not None:
            attrs['max'] = float(max)
        if min is not None:
            attrs['min'] = float(min)

        if container.target_opset < 6:
            attrs['consumed_inputs'] = [0]
            op_version = 1
        else:
            op_version = 6

        container.add_node(
            'Clip', input_name, output_name, op_version=op_version, **attrs)
    else:
        if container.target_opset < 12:
            op_version = 11
        else:
            op_version = 12
        if min is None and max is not None:
            raise RuntimeError(
                "Operator 'Clip': min must be specified if max is.")
        inputs = [input_name]

        if min is not None:
            if isinstance(min, (np.ndarray, float, int, np.float32,
                                np.float64, np.int64, np.int32)):
                # add initializer
                if isinstance(min, np.ndarray):
                    if len(min.shape) == 0:
                        min = [min]
                    elif min.shape == (1,):
                        min = (
                            list(min[0]) if hasattr(min[0], '__iter__')
                            else list(min))
                    else:
                        raise RuntimeError(
                            "min must be an array of one element.")
                else:
                    min = [min]

                # container in sklearn-onnx stores the computation type in
                # container.dtype.
                min_name = scope.get_unique_variable_name('clip_min')
                if op_version < 12:
                    min = np.array(
                        min, dtype=getattr(container, 'dtype', np.float32))
                    container.add_initializer(
                        min_name, getattr(container, 'proto_dtype',
                                          onnx_proto.TensorProto.FLOAT),
                        [], [min[0]])
                else:
                    min = np.array(min)
                    container.add_initializer(
                        min_name, NP_TYPE_TO_TENSOR_TYPE[min.dtype],
                        [], [min[0]])
                min = min_name
            if isinstance(min, str):
                inputs.append(min)
            else:
                raise RuntimeError(
                    "Parameter 'min' must be a string or a float.")

        if max is not None:
            if min is None:
                raise RuntimeError(
                    "Parameter 'min' must be specified if 'max' is.")
            if isinstance(max, (np.ndarray, float, int, np.float32,
                                np.float64, np.int64, np.int32)):
                # add initializer
                if isinstance(max, np.ndarray):
                    if len(max.shape) == 0:
                        max = [max]
                    elif max.shape == (1,):
                        max = (
                            list(max[0]) if hasattr(max[0], '__iter__')
                            else list(max))
                    else:
                        raise RuntimeError(
                            "max must be an array of one element.")
                else:
                    max = [max]

                max_name = scope.get_unique_variable_name('clip_max')
                if op_version < 12:
                    max = np.array(
                        max, dtype=getattr(container, 'dtype', np.float32))
                    container.add_initializer(
                        max_name, getattr(container, 'proto_dtype',
                                          onnx_proto.TensorProto.FLOAT),
                        [], [max[0]])
                else:
                    max = np.array(max)
                    container.add_initializer(
                        max_name, NP_TYPE_TO_TENSOR_TYPE[max.dtype],
                        [], [max[0]])
                max = max_name
            if isinstance(max, str):
                inputs.append(max)
            else:
                raise RuntimeError(
                    "Parameter 'max' must be a string or a float.")

        container.add_node(
            'Clip', inputs, output_name, op_version=op_version, **attrs)



# SPDX-License-Identifier: Apache-2.0

from logging import getLogger
from io import BytesIO
import numpy as np
import onnx  # noqa
from onnx import shape_inference, TensorProto
from onnx.numpy_helper import from_array, to_array
from onnx.helper import make_tensor
from ..proto.onnx_helper_modified import (
    make_node, make_tensor_value_info, make_graph,
    make_model, ValueInfoProto
)
from ..proto import get_latest_tested_opset_version
from onnx import onnx_pb as onnx_proto
from ..common._topology import Variable


def load_onnx_model(onnx_file_or_bytes):
    """
    Loads an *ONNX* file.

    :param onnx_file_or_bytes: *ONNX* file or bytes
    :return: *ONNX* model
    """
    if isinstance(onnx_file_or_bytes, str):
        with open(onnx_file_or_bytes, "rb") as f:
            return onnx.load(f)
    elif hasattr(onnx_file_or_bytes, 'read'):
        return onnx.load(onnx_file_or_bytes)
    else:
        b = BytesIO(onnx_file_or_bytes)
        return onnx.load(b)


def save_onnx_model(model, filename=None):
    """
    Saves a model as a file or bytes.

    :param model: *ONNX* model
    :param filename: filename or None to return bytes
    :return: bytes
    """
    content = model.SerializeToString()
    if filename is not None:
        if hasattr(filename, 'write'):
            filename.write(content)
        else:
            with open(filename, "wb") as f:
                f.write(content)
    return content


def enumerate_model_node_outputs(model, add_node=False):
    """
    Enumerates all the nodes of a model.

    :param model: ONNX graph
    :param add_node: if False, the function enumerates
        all output names from every node, otherwise, it
        enumerates tuple (output name, node)
    :return: enumerator
    """
    if not hasattr(model, "graph"):
        raise TypeError("Parameter model is not an ONNX model but "
                        "{}".format(type(model)))
    for node in model.graph.node:
        for out in node.output:
            yield (out, node) if add_node else out


def enumerate_model_initializers(model, add_node=False):
    """
    Enumerates all the initializers of a model.

    :param model: ONNX graph
    :param add_node: if False, the function enumerates
        all output names from every node, otherwise, it
        enumerates tuple (output name, node)
    :return: enumerator
    """
    for node in model.graph.initializer:
        yield (node.name, node) if add_node else node.name


def select_model_inputs_outputs(model, outputs=None, inputs=None):
    """
    Takes a model and changes its outputs.

    :param model: *ONNX* model
    :param inputs: new inputs
    :param outputs: new outputs
    :return: modified model

    The function removes unneeded files.
    """
    if inputs is not None:
        raise NotImplementedError("Parameter inputs cannot be empty.")
    if outputs is None:
        raise RuntimeError("Parameter outputs cannot be None.")
    if not isinstance(outputs, list):
        outputs = [outputs]

    mark_var = {}
    for out in enumerate_model_node_outputs(model):
        mark_var[out] = 0
    for inp in model.graph.input:
        mark_var[inp.name] = 0
    for out in outputs:
        if out not in mark_var:
            raise ValueError("Output '{}' not found in model.".format(out))
        mark_var[out] = 1

    nodes = model.graph.node[::-1]
    mark_op = {}
    for node in nodes:
        mark_op[node.name] = 0

    # We mark all the nodes we need to keep.
    nb = 1
    while nb > 0:
        nb = 0
        for node in nodes:
            if mark_op[node.name] == 1:
                continue
            mod = False
            for out in node.output:
                if mark_var[out] == 1:
                    mark_op[node.name] = 1
                    mod = True
                    break
            if not mod:
                continue

            nb += 1
            for inp in node.input:
                if mark_var.get(inp, 0) == 1:
                    continue
                mark_var[inp] = 1
                nb += 1

    # All nodes verifies mark_op[node.name] == 1
    keep_nodes = [node for node in nodes if mark_op[node.name] == 1]

    var_out = []
    for out in outputs:
        value_info = ValueInfoProto()
        value_info.name = out
        var_out.append(value_info)
    graph = make_graph(keep_nodes, model.graph.name, model.graph.input,
                       var_out, model.graph.initializer)
    onnx_model = make_model(graph)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:
        values = {p.key: p.value for p in model.metadata_props}
        onnx.helper.set_model_props(onnx_model, values)

    if len(onnx_model.graph.input) != len(model.graph.input):
        raise RuntimeError("Input mismatch {} != {}".format(
            len(onnx_model.input), len(model.input)))

    # fix opset import
    del onnx_model.opset_import[:]
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()
        op_set.domain = oimp.domain
        op_set.version = oimp.version
    return onnx_model


def infer_outputs(op_type, inputs, outputs=None, initializer=None,
                  target_opset=None, **atts):
    """
    Infers outputs type and shapes given an ONNX operator.
    """
    logger = getLogger('skl2onnx')
    logger.debug(
        '[infer_outputs] op_type=%r inputs=%r outputs=%r',
        op_type, [x.name for x in inputs], outputs)
    if isinstance(op_type, str):
        required_outputs = []
        if outputs:
            for o in outputs:
                if hasattr(o, 'onnx_name'):
                    required_outputs.append(o.onnx_name)
                elif isinstance(o, str):
                    required_outputs.append(o)
                else:
                    raise TypeError("Unable to require output {}.".format(o))
        node = make_node(op_type, [i.onnx_name for i in inputs],
                         required_outputs, **atts)
        node = [node]
    elif hasattr(op_type, 'nodes'):
        node = op_type.nodes
    else:
        raise RuntimeError("Unable to build ONNX nodes from type {}.".format(
            type(op_type)))

    input_init = inputs.copy()
    if initializer:
        input_init.extend(initializer)
    onnx_inputs = []
    for input in input_init:
        if isinstance(input, Variable):
            onnx_type = input.type.to_onnx_type()
            tensor_type = onnx_type.tensor_type
            shape = [tensor_type.shape.dim[i].dim_value
                     for i in range(len(tensor_type.shape.dim))]
            inp = make_tensor_value_info(input.onnx_name,
                                         tensor_type.elem_type,
                                         tuple(shape))
            onnx_inputs.append(inp)
        elif isinstance(input, onnx.TensorProto):
            v = make_tensor_value_info(
                input.name, input.data_type.real,
                list(d for d in input.dims))
            onnx_inputs.append(v)
        elif isinstance(input, onnx.AttributeProto):
            value_info = ValueInfoProto()
            value_info.name = input.name
            onnx_type = onnx_proto.TypeProto()
            onnx_type.tensor_type.elem_type = input.type
            value_info.type.CopyFrom(onnx_type)
            onnx_inputs.append(value_info)
        else:
            onnx_inputs.append(input)

    graph = make_graph(node, 'infer_shapes',
                       onnx_inputs, [])
    original_model = make_model(graph, producer_name='skl2onnx')
    domains = {}
    for n in node:
        domains[n.domain] = max(domains.get(n.domain, 1),
                                getattr(n, 'op_version', 1))
    for i, (k, v) in enumerate(domains.items()):
        if i == 0 and len(original_model.opset_import) == 1:
            op_set = original_model.opset_import[0]
        else:
            op_set = original_model.opset_import.add()
        op_set.domain = k
        if target_opset:
            if isinstance(target_opset, dict):
                op_set.version = target_opset.get(
                    k, get_latest_tested_opset_version())
            else:
                op_set.version = target_opset
        else:
            op_set.version = get_latest_tested_opset_version()

    try:
        inferred_model = shape_inference.infer_shapes(original_model)
    except RuntimeError as e:
        raise RuntimeError(
            "Unable to infer shape of node '{}'\n{}".format(
                op_type, original_model)) from e
    all_shapes = Variable.from_pb(inferred_model.graph.value_info)
    used = set()
    for node in graph.node:
        for name in node.input:
            used.add(name)
    shapes = [shape for shape in all_shapes if shape.onnx_name not in used]
    if len(shapes) == 0:
        raise RuntimeError("Shape inference fails.\n"
                           "*Inputs*\n{}\n*Model*\n{}'".format(
                               onnx_inputs, original_model))
    logger.debug('[infer_outputs] shapes=%r', shapes)
    return shapes


def change_onnx_domain(model, ops):
    """
    Takes a model and changes its outputs.

    :param model: *ONNX* model
    :param ops: dictionary { optype: ('optype', 'new domain') }
    :return: modified model

    The function removes unneeded files.
    """
    nodes = model.graph.node
    for node in nodes:
        rep = ops.get(node.op_type, None)
        if rep is None:
            continue
        node.op_type = rep[0]
        node.domain = rep[1]

    graph = make_graph(nodes, model.graph.name, model.graph.input,
                       model.graph.output, model.graph.initializer)
    onnx_model = make_model(graph)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:
        values = {p.key: p.value for p in model.metadata_props}
        onnx.helper.set_model_props(onnx_model, values)

    if len(onnx_model.graph.input) != len(model.graph.input):
        raise RuntimeError("Input mismatch {} != {}".format(
            len(onnx_model.input), len(model.input)))

    # fix opset import
    domain_set = set()
    has_domain = False
    del onnx_model.opset_import[:]
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()
        op_set.domain = oimp.domain
        op_set.version = oimp.version
        domain_set.add(oimp.domain)
        if not has_domain:
            has_domain = oimp.domain in domain_set
    for v in ops.values():
        if v[1] not in domain_set:
            op_set = onnx_model.opset_import.add()
            op_set.domain = v[1]
            op_set.version = 1
    return onnx_model


def add_output_initializer(model_onnx, name, value, suffix='_init'):
    """
    Add a constant and link it to one output.
    It allows the user to store arrays into the graph
    and retrieve them when using it.
    The initializer is named `name + suffix`, the output
    is named `name`.

    :param model_onnx: ONNX graph
    :param name: initializer name (initializer name, output name)
    :param value: array to store
    :param suffix: name of the initializer
    :return: new model

    It is possible to add multiple constant by using list:
    ``add_output_initializer(model_onnx, ['name1', 'name2'], [v1, v2])``.
    """
    if isinstance(name, str):
        name_list = [name]
        value_list = [value]
    else:
        name_list = name
        value_list = value

    if len(name_list) != len(value_list):
        raise ValueError(
            "Mismatched names and values. There are %d names and %d values."
            "" % (len(name_list), len(value_list)))

    nodes = list(model_onnx.graph.node)
    inits = list(model_onnx.graph.initializer)
    outputs = list(model_onnx.graph.output)

    for name, value in zip(name_list, value_list):
        name_output = name
        name_init = name + suffix
        names = set(i.name for i in model_onnx.graph.initializer)
        if name_output in names or name_init in names:
            raise ValueError(
                "Names %r or %r is already taken by an initializer: %r." % (
                    name_output, name_init, ", ".join(sorted(names))))
        names = set(i.name for i in model_onnx.graph.output)
        if name_output in names or name_init in names:
            raise ValueError(
                "Names %r or %r is already taken by an output: %r." % (
                    name_output, name_init, ", ".join(sorted(names))))
        names = set(i.name for i in model_onnx.graph.input)
        if name_output in names or name_init in names:
            raise ValueError(
                "Names %r or %r is already taken by an output: %r." % (
                    name_output, name_init, ", ".join(sorted(names))))

        try:
            cst = from_array(value, name=name_init)
        except RuntimeError as e:
            st = str(value.dtype).lower()
            if st.startswith('u') or st.startswith("<u"):
                cst_value = np.array([s.encode('utf-8') for s in value])
                cst = make_tensor(
                    name_init, data_type=TensorProto.STRING,
                    dims=value.shape, vals=list(cst_value))
            else:
                raise e

        inits.append(cst)

        outputs.append(make_tensor_value_info(
            name_output, cst.data_type, cst.dims))

        nodes.append(make_node('Identity', [name_init], [name_output]))

    graph = make_graph(
            nodes, model_onnx.graph.name, model_onnx.graph.input,
            outputs, inits)

    onnx_model = make_model(graph)
    onnx_model.ir_version = model_onnx.ir_version
    onnx_model.producer_name = model_onnx.producer_name
    onnx_model.producer_version = model_onnx.producer_version
    onnx_model.domain = model_onnx.domain
    onnx_model.model_version = model_onnx.model_version
    onnx_model.doc_string = model_onnx.doc_string
    if len(model_onnx.metadata_props) > 0:
        values = {p.key: p.value for p in model_onnx.metadata_props}
        onnx.helper.set_model_props(onnx_model, values)

    if len(onnx_model.graph.input) != len(model_onnx.graph.input):
        raise RuntimeError("Input mismatch {} != {}".format(
            len(onnx_model.input), len(model_onnx.input)))

    # fix opset import
    del onnx_model.opset_import[:]
    for oimp in model_onnx.opset_import:
        op_set = onnx_model.opset_import.add()
        op_set.domain = oimp.domain
        op_set.version = oimp.version
    return onnx_model


def get_initializers(model_onnx):
    """
    Retrieves the list of initializers in a model in a
    dictionary `{ name: value }`.
    """
    res = {}
    for init in model_onnx.graph.initializer:
        res[init.name] = to_array(init)
    return res


def update_onnx_initializers(model_onnx, new_inits):
    """
    Updates initializer in a ONNX model.

    :param model_onnx: ONNX model
    :param new_inits: new initializers
    :return: list of updated initializers
    """
    updated = []
    replace_weights = []
    replace_indices = []
    for i, w in enumerate(model_onnx.graph.initializer):
        if w.name in new_inits:
            replace_weights.append(from_array(new_inits[w.name], w.name))
            replace_indices.append(i)
            updated.append(w.name)
    replace_indices.sort(reverse=True)
    for w_i in replace_indices:
        del model_onnx.graph.initializer[w_i]
    model_onnx.graph.initializer.extend(replace_weights)
    return updated



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.shape_calculator import calculate_linear_classifier_output_shapes


register_shape_calculator('SklearnLinearClassifier',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnLinearSVC',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnAdaBoostClassifier',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnBaggingClassifier',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnBernoulliNB',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnCategoricalNB',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnComplementNB',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnGaussianNB',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnMultinomialNB',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnCalibratedClassifierCV',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnMLPClassifier',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnSGDClassifier',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnStackingClassifier',
                          calculate_linear_classifier_output_shapes)



# SPDX-License-Identifier: Apache-2.0


import copy
import numpy as np
from ..common._registration import register_shape_calculator
from ..common.shape_calculator import calculate_linear_classifier_output_shapes
from ..common.data_types import (
    FloatTensorType, Int64TensorType, DoubleTensorType
)
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_neighbours_transformer(operator):
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[
            FloatTensorType, Int64TensorType, DoubleTensorType])

    N = operator.inputs[0].get_first_dimension()
    n_samples_fit = operator.raw_operator.n_samples_fit_
    output_type = (
        DoubleTensorType
        if isinstance(operator.inputs[0].type, DoubleTensorType)
        else FloatTensorType
    )
    operator.outputs[0].type = output_type([N, n_samples_fit])


def calculate_sklearn_nearest_neighbours(operator):
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=[1, 2])
    check_input_and_output_types(
        operator, good_input_types=[
            FloatTensorType, Int64TensorType, DoubleTensorType])

    N = operator.inputs[0].get_first_dimension()
    neighbours = operator.raw_operator.n_neighbors
    operator.outputs[0].type = Int64TensorType([N, neighbours])
    operator.outputs[1].type.shape = [N, neighbours]


def calculate_sklearn_nearest_neighbours_regressor(operator):
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=[1, 2])
    check_input_and_output_types(
        operator, good_input_types=[
            FloatTensorType, Int64TensorType, DoubleTensorType])

    N = operator.inputs[0].get_first_dimension()
    if (hasattr(operator.raw_operator, '_y') and
            len(np.squeeze(operator.raw_operator._y).shape) == 1):
        C = 1
    else:
        C = operator.raw_operator._y.shape[-1]
    operator.outputs[0].type.shape = [N, C]


def calculate_sklearn_nca(operator):
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[
            FloatTensorType, Int64TensorType, DoubleTensorType])

    N = operator.inputs[0].get_first_dimension()
    output_type = (
        DoubleTensorType
        if isinstance(operator.inputs[0].type, DoubleTensorType)
        else FloatTensorType
    )
    n_components = operator.raw_operator.components_.shape[0]
    operator.outputs[0].type = output_type([N, n_components])


def calculate_sklearn_knn_imputer(operator):
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType])

    operator.outputs[0].type = copy.deepcopy(operator.inputs[0].type)
    operator.outputs[0].type.shape = operator.inputs[0].type.shape


register_shape_calculator('SklearnKNeighborsRegressor',
                          calculate_sklearn_nearest_neighbours_regressor)
register_shape_calculator('SklearnRadiusNeighborsRegressor',
                          calculate_sklearn_nearest_neighbours_regressor)
register_shape_calculator('SklearnKNeighborsClassifier',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnRadiusNeighborsClassifier',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnKNNImputer',
                          calculate_sklearn_knn_imputer)
register_shape_calculator('SklearnKNeighborsTransformer',
                          calculate_sklearn_neighbours_transformer)
register_shape_calculator('SklearnNearestNeighbors',
                          calculate_sklearn_nearest_neighbours)
register_shape_calculator(
    'SklearnNeighborhoodComponentsAnalysis', calculate_sklearn_nca)



# SPDX-License-Identifier: Apache-2.0

"""
Place holder for all ONNX operators.
"""
import sys
import textwrap
from sklearn.pipeline import Pipeline, FeatureUnion
try:
    from sklearn.compose import ColumnTransformer
except ImportError:
    # ColumnTransformer was introduced in 0.20.
    ColumnTransformer = None
from .onnx_subgraph_operator_mixin import OnnxSubGraphOperatorMixin


def ClassFactorySklearn(skl_obj, class_name, doc, conv, shape_calc, alias):
    from .onnx_subgraph_operator_mixin import OnnxSubGraphOperatorMixin

    newclass = type(class_name, (OnnxSubGraphOperatorMixin, skl_obj),
                    {'__doc__': doc,
                     'operator_name': skl_obj.__name__,
                     '_fct_converter': conv,
                     '_fct_shape_calc': shape_calc,
                     'input_range': [1, 1e9],
                     'output_range': [1, 1e9],
                     'op_version': None,
                     'alias': alias,
                     '__module__': __name__})
    return newclass


def dynamic_class_creation_sklearn():
    """
    Automatically generates classes for each of the converter.
    """
    from ..common._registration import _shape_calculator_pool, _converter_pool
    from .._supported_operators import sklearn_operator_name_map

    cls = {}

    for skl_obj, name in sklearn_operator_name_map.items():
        if skl_obj is None:
            continue
        conv = _converter_pool[name]
        shape_calc = _shape_calculator_pool[name]
        skl_name = skl_obj.__name__
        doc = ["OnnxOperatorMixin for **{}**".format(skl_name), ""]
        if conv.__doc__:
            doc.append(textwrap.dedent(conv.__doc__))
        doc = "\n".join(doc)
        prefix = "Sklearn" if "sklearn" in str(skl_obj) else ""
        class_name = "Onnx" + prefix + skl_name
        try:
            cl = ClassFactorySklearn(skl_obj, class_name,
                                     doc, conv, shape_calc,
                                     name)
        except TypeError:
            continue
        cls[class_name] = cl
    return cls


def _update_module():
    """
    Dynamically updates the module with operators defined
    by *ONNX*.
    """
    res = dynamic_class_creation_sklearn()
    this = sys.modules[__name__]
    for k, v in res.items():
        setattr(this, k, v)


def find_class(skl_cl):
    """
    Finds the corresponding :class:`OnnxSubGraphOperatorMixin`
    class to *skl_cl*.
    """
    name = skl_cl.__name__
    prefix = "OnnxSklearn"
    full_name = prefix + name
    this = sys.modules[__name__]
    if not hasattr(this, full_name):
        available = sorted(filter(lambda n: prefix in n, sys.modules))
        raise RuntimeError(
            "Unable to find a class for '{}' in\n{}".format(
                skl_cl.__name__, "\n".join(available)))
    cl = getattr(this, full_name)
    if "automation" in str(cl):
        raise RuntimeError("Dynamic operation issue with class "
                           "name '{}' from '{}'.".format(cl, __name__))
    return cl


class OnnxSklearnPipeline(Pipeline, OnnxSubGraphOperatorMixin):
    """
    Combines `Pipeline
    <https://scikit-learn.org/stable/modules/generated/
    sklearn.pipeline.Pipeline.html>`_ and
    :class:`OnnxSubGraphOperatorMixin`.
    """

    def __init__(self, steps, memory=None, verbose=False, op_version=None):
        Pipeline.__init__(self, steps=steps, memory=memory, verbose=verbose)
        OnnxSubGraphOperatorMixin.__init__(self)
        self.op_version = op_version


if ColumnTransformer is not None:

    class OnnxSklearnColumnTransformer(ColumnTransformer,
                                       OnnxSubGraphOperatorMixin):
        """
        Combines `ColumnTransformer
        <https://scikit-learn.org/stable/modules/generated/
        sklearn.compose.ColumnTransformer.html>`_ and
        :class:`OnnxSubGraphOperatorMixin`.
        """

        def __init__(self, op_version=None):
            self.op_version = op_version


class OnnxSklearnFeatureUnion(FeatureUnion, OnnxSubGraphOperatorMixin):
    """
    Combines `FeatureUnion
    <https://scikit-learn.org/stable/modules/generated/
    sklearn.pipeline.FeatureUnion.html>`_ and
    :class:`OnnxSubGraphOperatorMixin`.
    """

    def __init__(self, op_version=None):
        self.op_version = op_version


_update_module()



# SPDX-License-Identifier: Apache-2.0


import numbers
import numpy as np
from sklearn.ensemble import RandomTreesEmbedding
from ..common._apply_operation import (
    apply_cast,
    apply_concat,
    apply_reshape,
    apply_transpose,
)
from ..common.data_types import (
    BooleanTensorType, Int64TensorType, guess_numpy_type)
from ..common._registration import register_converter
from ..common.tree_ensemble import (
    add_tree_to_attribute_pairs,
    add_tree_to_attribute_pairs_hist_gradient_boosting,
    get_default_tree_classifier_attribute_pairs,
    get_default_tree_regressor_attribute_pairs
)
from ..common.utils_classifier import get_label_classes
from ..proto import onnx_proto
from .decision_tree import (
    predict, _build_labels_path, _build_labels_leaf, _append_decision_output)


def _num_estimators(op):
    # don't use op.n_estimators since it may not be the same as
    # len(op.estimators_). At training time n_estimators can be changed by
    # training code:
    #   for j in range(10):
    #       ...
    #       classifier.fit(X_tmp, y_tmp)
    #       classifier.n_estimators += 30
    if hasattr(op, 'estimators_'):
        return len(op.estimators_)
    elif hasattr(op, '_predictors'):
        # HistGradientBoosting*
        return len(op._predictors)
    raise NotImplementedError(
        "Model should have attribute 'estimators_' or '_predictors'.")


def _calculate_labels(scope, container, model, proba):
    predictions = []
    transposed_result_name = scope.get_unique_variable_name(
        'transposed_result')
    apply_transpose(scope, proba, transposed_result_name,
                    container, perm=(1, 2, 0))
    for k in range(model.n_outputs_):
        preds_name = scope.get_unique_variable_name('preds')
        reshaped_preds_name = scope.get_unique_variable_name(
            'reshaped_preds')
        k_name = scope.get_unique_variable_name('k_column')
        out_k_name = scope.get_unique_variable_name('out_k_column')
        argmax_output_name = scope.get_unique_variable_name(
            'argmax_output')
        classes_name = scope.get_unique_variable_name('classes')
        reshaped_result_name = scope.get_unique_variable_name(
            'reshaped_result')

        container.add_initializer(
            k_name, onnx_proto.TensorProto.INT64,
            [], [k])
        container.add_initializer(
            classes_name, onnx_proto.TensorProto.INT64,
            model.classes_[k].shape, model.classes_[k])

        container.add_node(
            'ArrayFeatureExtractor', [transposed_result_name, k_name],
            out_k_name, op_domain='ai.onnx.ml',
            name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
        container.add_node(
            'ArgMax', out_k_name, argmax_output_name,
            name=scope.get_unique_operator_name('ArgMax'), axis=1)
        apply_reshape(scope, argmax_output_name, reshaped_result_name,
                      container, desired_shape=(1, -1))
        container.add_node(
            'ArrayFeatureExtractor', [classes_name, reshaped_result_name],
            preds_name, op_domain='ai.onnx.ml',
            name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
        apply_reshape(scope, preds_name, reshaped_preds_name,
                      container, desired_shape=(-1, 1))
        predictions.append(reshaped_preds_name)
    return predictions


def convert_sklearn_random_forest_classifier(
        scope, operator, container, op_type='TreeEnsembleClassifier',
        op_domain='ai.onnx.ml', op_version=1):
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    attr_dtype = dtype if op_version >= 3 else np.float32
    op = operator.raw_operator

    if hasattr(op, 'n_outputs_'):
        n_outputs = int(op.n_outputs_)
        options = container.get_options(
            op, dict(raw_scores=False, decision_path=False,
                     decision_leaf=False))
    elif hasattr(op, 'n_trees_per_iteration_'):
        # HistGradientBoostingClassifier
        n_outputs = op.n_trees_per_iteration_
        options = container.get_options(op, dict(raw_scores=False))
    else:
        raise NotImplementedError(
            "Model should have attribute 'n_outputs_' or "
            "'n_trees_per_iteration_'.")

    use_raw_scores = options['raw_scores']

    if n_outputs == 1 or hasattr(op, 'loss_') or hasattr(op, '_loss'):
        classes = get_label_classes(scope, op)

        if all(isinstance(i, np.ndarray) for i in classes):
            classes = np.concatenate(classes)
        attr_pairs = get_default_tree_classifier_attribute_pairs()
        attr_pairs['name'] = scope.get_unique_operator_name(op_type)

        if all(isinstance(i, (numbers.Real, bool, np.bool_)) for i in classes):
            class_labels = [int(i) for i in classes]
            attr_pairs['classlabels_int64s'] = class_labels
        elif all(isinstance(i, str) for i in classes):
            class_labels = [str(i) for i in classes]
            attr_pairs['classlabels_strings'] = class_labels
        else:
            raise ValueError(
                'Only string and integer class labels are allowed.')

        # random forest calculate the final score by averaging over all trees'
        # outcomes, so all trees' weights are identical.
        if hasattr(op, 'estimators_'):
            estimator_count = len(op.estimators_)
            tree_weight = 1. / estimator_count
        elif hasattr(op, '_predictors'):
            # HistGradientBoostingRegressor
            estimator_count = len(op._predictors)
            tree_weight = 1.
        else:
            raise NotImplementedError(
                "Model should have attribute 'estimators_' or '_predictors'.")

        for tree_id in range(estimator_count):

            if hasattr(op, 'estimators_'):
                tree = op.estimators_[tree_id].tree_
                add_tree_to_attribute_pairs(
                    attr_pairs, True, tree, tree_id,
                    tree_weight, 0, True, True,
                    dtype=dtype)
            else:
                # HistGradientBoostClassifier
                if len(op._predictors[tree_id]) == 1:
                    tree = op._predictors[tree_id][0]
                    add_tree_to_attribute_pairs_hist_gradient_boosting(
                        attr_pairs, True, tree, tree_id, tree_weight, 0,
                        False, False, dtype=dtype)
                else:
                    for cl, tree in enumerate(op._predictors[tree_id]):
                        add_tree_to_attribute_pairs_hist_gradient_boosting(
                            attr_pairs, True, tree, tree_id * n_outputs + cl,
                            tree_weight, cl, False, False,
                            dtype=dtype)

        if hasattr(op, '_baseline_prediction'):
            if isinstance(op._baseline_prediction, np.ndarray):
                attr_pairs['base_values'] = list(
                    op._baseline_prediction.ravel())
            else:
                attr_pairs['base_values'] = [op._baseline_prediction]

        if hasattr(op, 'loss_'):
            loss = op.loss_
        elif hasattr(op, '_loss'):
            # scikit-learn >= 0.24
            loss = op._loss
        else:
            loss = None
        if loss is not None:
            if use_raw_scores:
                attr_pairs['post_transform'] = "NONE"
            elif loss.__class__.__name__ in (
                    "BinaryCrossEntropy", "HalfBinomialLoss"):
                attr_pairs['post_transform'] = "LOGISTIC"
            elif loss.__class__.__name__ in (
                    "CategoricalCrossEntropy", "HalfMultinomialLoss"):
                attr_pairs['post_transform'] = "SOFTMAX"
            else:
                raise NotImplementedError(
                    "There is no corresponding post_transform for "
                    "'{}'.".format(loss.__class__.__name__))
        elif use_raw_scores:
            raise RuntimeError(
                "The converter cannot implement decision_function for "
                "'{}' and loss '{}'.".format(type(op), loss))

        input_name = operator.input_full_names
        if type(operator.inputs[0].type) == BooleanTensorType:
            cast_input_name = scope.get_unique_variable_name('cast_input')

            apply_cast(scope, input_name, cast_input_name,
                       container, to=onnx_proto.TensorProto.FLOAT)
            input_name = cast_input_name

        if dtype is not None:
            for k in attr_pairs:
                if k in ('nodes_values', 'class_weights',
                         'target_weights', 'nodes_hitrates',
                         'base_values'):
                    attr_pairs[k] = np.array(
                        attr_pairs[k], dtype=attr_dtype).ravel()

        container.add_node(
            op_type, input_name,
            [operator.outputs[0].full_name, operator.outputs[1].full_name],
            op_domain=op_domain, op_version=op_version, **attr_pairs)

        if (not options.get('decision_path', False) and
                not options.get('decision_leaf', False)):
            return

        # decision_path
        tree_paths = []
        tree_leaves = []
        for i, tree in enumerate(op.estimators_):

            attrs = get_default_tree_classifier_attribute_pairs()
            attrs['name'] = scope.get_unique_operator_name(
                "%s_%d" % (op_type, i))
            attrs['n_targets'] = int(op.n_outputs_)
            add_tree_to_attribute_pairs(
                attrs, True, tree.tree_, 0, 1., 0, False,
                True, dtype=dtype)

            attrs['n_targets'] = 1
            attrs['post_transform'] = 'NONE'
            attrs['target_ids'] = [0 for _ in attrs['class_ids']]
            attrs['target_weights'] = [
                float(_) for _ in attrs['class_nodeids']]
            attrs['target_nodeids'] = attrs['class_nodeids']
            attrs['target_treeids'] = attrs['class_treeids']
            rem = [k for k in attrs if k.startswith('class')]
            for k in rem:
                del attrs[k]

            if dtype is not None:
                for k in attrs:
                    if k in ('nodes_values', 'class_weights',
                             'target_weights', 'nodes_hitrates',
                             'base_values'):
                        attrs[k] = np.array(attrs[k], dtype=attr_dtype).ravel()

            if options['decision_path']:
                # decision_path
                tree_paths.append(
                    _append_decision_output(
                        input_name, attrs, _build_labels_path, None,
                        scope, operator, container,
                        op_type=op_type, op_domain=op_domain,
                        op_version=op_version, regression=True,
                        overwrite_tree=tree.tree_))
            if options['decision_leaf']:
                # decision_path
                tree_leaves.append(
                    _append_decision_output(
                        input_name, attrs, _build_labels_leaf, None,
                        scope, operator, container,
                        op_type=op_type, op_domain=op_domain,
                        op_version=op_version, regression=True,
                        cast_encode=True))

        # merges everything
        n_out = 2
        if options['decision_path']:
            apply_concat(
                scope, tree_paths, operator.outputs[n_out].full_name,
                container, axis=1,
                operator_name=scope.get_unique_operator_name('concat'))
            n_out += 1

        if options['decision_leaf']:
            # decision_path
            apply_concat(
                scope, tree_leaves, operator.outputs[n_out].full_name,
                container, axis=1,
                operator_name=scope.get_unique_operator_name('concat'))
            n_out += 1

    else:
        if use_raw_scores:
            raise RuntimeError(
                "The converter cannot implement decision_function for "
                "'{}'.".format(type(op)))
        concatenated_proba_name = scope.get_unique_variable_name(
            'concatenated_proba')
        proba = []
        for est in op.estimators_:
            reshaped_est_proba_name = scope.get_unique_variable_name(
                'reshaped_est_proba')
            est_proba = predict(
                est, scope, operator, container, op_type, op_domain,
                op_version, is_ensemble=True)
            apply_reshape(
                scope, est_proba, reshaped_est_proba_name, container,
                desired_shape=(
                    1, n_outputs, -1, max([len(x) for x in op.classes_])))
            proba.append(reshaped_est_proba_name)
        apply_concat(scope, proba, concatenated_proba_name,
                     container, axis=0)
        container.add_node('ReduceMean', concatenated_proba_name,
                           operator.outputs[1].full_name,
                           name=scope.get_unique_operator_name('ReduceMean'),
                           axes=[0], keepdims=0)
        predictions = _calculate_labels(
            scope, container, op, operator.outputs[1].full_name)
        apply_concat(scope, predictions, operator.outputs[0].full_name,
                     container, axis=1)

        if (options.get('decision_path', False) or
                options.get('decision_leaf', False)):
            raise RuntimeError(
                "Decision output for multi-outputs is not implemented yet.")


def convert_sklearn_random_forest_regressor_converter(
        scope, operator, container, op_type='TreeEnsembleRegressor',
        op_domain='ai.onnx.ml', op_version=1):
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    op = operator.raw_operator
    attrs = get_default_tree_regressor_attribute_pairs()
    attrs['name'] = scope.get_unique_operator_name(op_type)

    if hasattr(op, 'n_outputs_'):
        attrs['n_targets'] = int(op.n_outputs_)
    elif hasattr(op, 'n_trees_per_iteration_'):
        # HistGradientBoostingRegressor
        attrs['n_targets'] = op.n_trees_per_iteration_
    else:
        raise NotImplementedError(
            "Model should have attribute 'n_outputs_' or "
            "'n_trees_per_iteration_'.")

    if hasattr(op, 'estimators_'):
        estimator_count = len(op.estimators_)
        tree_weight = 1. / estimator_count
    elif hasattr(op, '_predictors'):
        # HistGradientBoostingRegressor
        estimator_count = len(op._predictors)
        tree_weight = 1.
    else:
        raise NotImplementedError(
            "Model should have attribute 'estimators_' or '_predictors'.")

    # random forest calculate the final score by averaging over all trees'
    # outcomes, so all trees' weights are identical.
    for tree_id in range(estimator_count):
        if hasattr(op, 'estimators_'):
            tree = op.estimators_[tree_id].tree_
            add_tree_to_attribute_pairs(attrs, False, tree, tree_id,
                                        tree_weight, 0, False, True,
                                        dtype=dtype)
        else:
            # HistGradientBoostingRegressor
            if len(op._predictors[tree_id]) != 1:
                raise NotImplementedError(
                    "The converter does not work when the number of trees "
                    "is not 1 but {}.".format(len(op._predictors[tree_id])))
            tree = op._predictors[tree_id][0]
            add_tree_to_attribute_pairs_hist_gradient_boosting(
                attrs, False, tree, tree_id, tree_weight, 0, False,
                False, dtype=dtype)

    if hasattr(op, '_baseline_prediction'):
        if isinstance(op._baseline_prediction, np.ndarray):
            attrs['base_values'] = list(op._baseline_prediction)
        else:
            attrs['base_values'] = [op._baseline_prediction]

    input_name = operator.input_full_names
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name('cast_input')

        apply_cast(scope, operator.input_full_names, cast_input_name,
                   container, to=onnx_proto.TensorProto.FLOAT)
        input_name = cast_input_name

    if dtype is not None:
        for k in attrs:
            if k in ('nodes_values', 'class_weights',
                     'target_weights', 'nodes_hitrates',
                     'base_values'):
                attrs[k] = np.array(attrs[k], dtype=dtype).ravel()

    container.add_node(
        op_type, input_name,
        operator.outputs[0].full_name, op_domain=op_domain,
        op_version=op_version, **attrs)

    if hasattr(op, 'n_trees_per_iteration_'):
        # HistGradientBoostingRegressor does not implement decision_path.
        return
    if isinstance(op, RandomTreesEmbedding):
        options = scope.get_options(op)
    else:
        options = scope.get_options(
            op, dict(decision_path=False, decision_leaf=False))

    if (not options.get('decision_path', False) and
            not options.get('decision_leaf', False)):
        return

    # decision_path
    tree_paths = []
    tree_leaves = []
    for i, tree in enumerate(op.estimators_):

        attrs = get_default_tree_regressor_attribute_pairs()
        attrs['name'] = scope.get_unique_operator_name("%s_%d" % (op_type, i))
        attrs['n_targets'] = int(op.n_outputs_)
        add_tree_to_attribute_pairs(attrs, False, tree.tree_, 0, 1., 0, False,
                                    True, dtype=dtype)

        attrs['n_targets'] = 1
        attrs['post_transform'] = 'NONE'
        attrs['target_ids'] = [0 for _ in attrs['target_ids']]
        attrs['target_weights'] = [float(_) for _ in attrs['target_nodeids']]

        if dtype is not None:
            for k in attrs:
                if k in ('nodes_values', 'class_weights',
                         'target_weights', 'nodes_hitrates',
                         'base_values'):
                    attrs[k] = np.array(attrs[k], dtype=dtype).ravel()

        if options.get('decision_path', False):
            # decision_path
            tree_paths.append(
                _append_decision_output(
                    input_name, attrs, _build_labels_path, None,
                    scope, operator, container,
                    op_type=op_type, op_domain=op_domain,
                    op_version=op_version, regression=True,
                    overwrite_tree=tree.tree_))
        if options.get('decision_leaf', False):
            # decision_path
            tree_leaves.append(
                _append_decision_output(
                    input_name, attrs, _build_labels_leaf, None,
                    scope, operator, container,
                    op_type=op_type, op_domain=op_domain,
                    op_version=op_version, regression=True, cast_encode=True))

    # merges everything
    n_out = 1
    if options.get('decision_path', False):
        apply_concat(
            scope, tree_paths, operator.outputs[n_out].full_name, container,
            axis=1, operator_name=scope.get_unique_operator_name('concat'))
        n_out += 1

    if options.get('decision_leaf', False):
        # decision_path
        apply_concat(
            scope, tree_leaves, operator.outputs[n_out].full_name, container,
            axis=1, operator_name=scope.get_unique_operator_name('concat'))
        n_out += 1


register_converter('SklearnRandomForestClassifier',
                   convert_sklearn_random_forest_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'raw_scores': [True, False],
                            'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'decision_path': [True, False],
                            'decision_leaf': [True, False]})
register_converter('SklearnRandomForestRegressor',
                   convert_sklearn_random_forest_regressor_converter,
                   options={'decision_path': [True, False],
                            'decision_leaf': [True, False]})
register_converter('SklearnExtraTreesClassifier',
                   convert_sklearn_random_forest_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'raw_scores': [True, False],
                            'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'decision_path': [True, False],
                            'decision_leaf': [True, False]})
register_converter('SklearnExtraTreesRegressor',
                   convert_sklearn_random_forest_regressor_converter,
                   options={'decision_path': [True, False],
                            'decision_leaf': [True, False]})
register_converter('SklearnHistGradientBoostingClassifier',
                   convert_sklearn_random_forest_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'raw_scores': [True, False],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})
register_converter('SklearnHistGradientBoostingRegressor',
                   convert_sklearn_random_forest_regressor_converter,
                   options={'zipmap': [True, False, 'columns'],
                            'raw_scores': [True, False],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})



# SPDX-License-Identifier: Apache-2.0

import pprint
import numpy as np
from onnx import TensorProto
from ..common._apply_operation import (
    apply_abs, apply_add, apply_cast, apply_concat, apply_clip,
    apply_div, apply_exp, apply_mul, apply_reshape, apply_sub)
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import (
    guess_numpy_type, Int64TensorType, guess_proto_type)
from ..common._registration import register_converter
from .._supported_operators import sklearn_operator_name_map
from sklearn.ensemble import RandomForestClassifier


def _handle_zeros(scope, container, concatenated_prob_name,
                  reduced_prob_name, n_classes, proto_type):
    """
    This function replaces 0s in concatenated_prob_name with 1s and
    0s in reduced_prob_name with n_classes.
    """
    cast_prob_name = scope.get_unique_variable_name('cast_prob')
    bool_not_cast_prob_name = scope.get_unique_variable_name(
        'bool_not_cast_prob')
    mask_name = scope.get_unique_variable_name('mask')
    masked_concatenated_prob_name = scope.get_unique_variable_name(
        'masked_concatenated_prob')
    n_classes_name = scope.get_unique_variable_name('n_classes')
    reduced_prob_mask_name = scope.get_unique_variable_name(
        'reduced_prob_mask')
    masked_reduced_prob_name = scope.get_unique_variable_name(
        'masked_reduced_prob')

    proto_type2 = proto_type
    if proto_type2 not in (TensorProto.FLOAT, TensorProto.DOUBLE):
        proto_type2 = TensorProto.FLOAT

    container.add_initializer(n_classes_name, proto_type2,
                              [], [n_classes])

    apply_cast(scope, reduced_prob_name, cast_prob_name, container,
               to=TensorProto.BOOL)
    container.add_node('Not', cast_prob_name,
                       bool_not_cast_prob_name,
                       name=scope.get_unique_operator_name('Not'))
    apply_cast(scope, bool_not_cast_prob_name, mask_name, container,
               to=proto_type2)
    apply_add(scope, [concatenated_prob_name, mask_name],
              masked_concatenated_prob_name, container, broadcast=1)
    apply_mul(scope, [mask_name, n_classes_name], reduced_prob_mask_name,
              container, broadcast=1)
    apply_add(scope, [reduced_prob_name, reduced_prob_mask_name],
              masked_reduced_prob_name, container, broadcast=0)
    return masked_concatenated_prob_name, masked_reduced_prob_name


def _transform_sigmoid(scope, container, model, df_col_name, k, proto_type):
    """
    Sigmoid Calibration method
    """
    a_name = scope.get_unique_variable_name('a')
    b_name = scope.get_unique_variable_name('b')
    a_df_prod_name = scope.get_unique_variable_name('a_df_prod')
    exp_parameter_name = scope.get_unique_variable_name(
        'exp_parameter')
    exp_result_name = scope.get_unique_variable_name('exp_result')
    unity_name = scope.get_unique_variable_name('unity')
    denominator_name = scope.get_unique_variable_name('denominator')
    sigmoid_predict_result_name = scope.get_unique_variable_name(
        'sigmoid_predict_result')

    proto_type2 = proto_type
    if proto_type2 not in (TensorProto.FLOAT, TensorProto.DOUBLE):
        proto_type2 = TensorProto.FLOAT

    if hasattr(model, 'calibrators_'):
        # scikit-learn<1.1
        calibrators = model.calibrators_
    elif hasattr(model, 'calibrators'):
        # scikit-learn>=1.1
        calibrators = model.calibrators
    else:
        raise AttributeError(
            "Unable to find attribute calibrators_ or "
            "calibrators, check the model was trained, "
            "type=%r." % type(model))

    container.add_initializer(a_name, proto_type2,
                              [], [calibrators[k].a_])
    container.add_initializer(b_name, proto_type2,
                              [], [calibrators[k].b_])
    container.add_initializer(unity_name, proto_type2, [], [1])

    apply_mul(scope, [a_name, df_col_name], a_df_prod_name, container,
              broadcast=0)
    apply_add(scope, [a_df_prod_name, b_name], exp_parameter_name,
              container, broadcast=0)
    apply_exp(scope, exp_parameter_name, exp_result_name, container)
    apply_add(scope, [unity_name, exp_result_name], denominator_name,
              container, broadcast=0)
    apply_div(scope, [unity_name, denominator_name],
              sigmoid_predict_result_name, container, broadcast=0)
    return sigmoid_predict_result_name


def _transform_isotonic(scope, container, model, T, k, dtype, proto_type):
    """
    Isotonic calibration method
    This function can only handle one instance at a time because
    ArrayFeatureExtractor can only extract based on the last axis,
    so we can't fetch different columns for different rows.
    """
    if hasattr(model, 'calibrators_'):
        # scikit-learn<1.1
        calibrators = model.calibrators_
    elif hasattr(model, 'calibrators'):
        # scikit-learn>=1.1
        calibrators = model.calibrators
    else:
        raise AttributeError(
            "Unable to find attribute calibrators_ or "
            "calibrators, check the model was trained, "
            "type=%r." % type(model))

    if calibrators[k].out_of_bounds == 'clip':
        clipped_df_name = scope.get_unique_variable_name('clipped_df')
        apply_clip(scope, T, clipped_df_name, container,
                   operator_name=scope.get_unique_operator_name('Clip'),
                   max=np.array(calibrators[k].X_max_, dtype=dtype),
                   min=np.array(calibrators[k].X_min_, dtype=dtype))
        T = clipped_df_name

    reshaped_df_name = scope.get_unique_variable_name('reshaped_df')
    calibrator_x_name = scope.get_unique_variable_name('calibrator_x')
    calibrator_y_name = scope.get_unique_variable_name('calibrator_y')
    distance_name = scope.get_unique_variable_name('distance')
    absolute_distance_name = scope.get_unique_variable_name(
        'absolute_distance')
    nearest_x_index_name = scope.get_unique_variable_name(
        'nearest_x_index')
    nearest_y_name = scope.get_unique_variable_name('nearest_y')

    if hasattr(calibrators[k], '_X_'):
        atX, atY = '_X_', '_y_'
    elif hasattr(calibrators[k], '_necessary_X_'):
        atX, atY = '_necessary_X_', '_necessary_y_'
    elif hasattr(calibrators[k], 'X_thresholds_'):
        atX, atY = 'X_thresholds_', 'y_thresholds_'
    else:
        raise AttributeError(
            "Unable to find attribute '_X_' or '_necessary_X_' "
            "for type {}\n{}."
            "".format(type(calibrators[k]),
                      pprint.pformat(dir(calibrators[k]))))

    proto_type2 = proto_type
    if proto_type2 not in (TensorProto.FLOAT, TensorProto.DOUBLE):
        proto_type2 = TensorProto.FLOAT

    container.add_initializer(
        calibrator_x_name, proto_type2,
        [len(getattr(calibrators[k], atX))],
        getattr(calibrators[k], atX))
    container.add_initializer(
        calibrator_y_name, proto_type2,
        [len(getattr(calibrators[k], atY))],
        getattr(calibrators[k], atY))

    apply_reshape(scope, T, reshaped_df_name, container,
                  desired_shape=(-1, 1))
    apply_sub(scope, [reshaped_df_name, calibrator_x_name],
              distance_name, container, broadcast=1)
    apply_abs(scope, distance_name, absolute_distance_name, container)
    container.add_node('ArgMin', absolute_distance_name,
                       nearest_x_index_name, axis=1,
                       name=scope.get_unique_operator_name('ArgMin'))
    container.add_node(
        'ArrayFeatureExtractor',
        [calibrator_y_name, nearest_x_index_name],
        nearest_y_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))

    nearest_y_name_reshaped = scope.get_unique_variable_name(
        'nearest_y_name_reshaped')
    apply_reshape(scope, nearest_y_name,
                  nearest_y_name_reshaped, container,
                  desired_shape=(-1, 1))
    return nearest_y_name_reshaped


def convert_calibrated_classifier_base_estimator(scope, operator, container,
                                                 model, model_index):
    # Computational graph:
    #
    # In the following graph, variable names are in lower case characters only
    # and operator names are in upper case characters. We borrow operator names
    # from the official ONNX spec:
    # https://github.com/onnx/onnx/blob/master/docs/Operators.md
    # All variables are followed by their shape in [].
    #
    # Symbols:
    # M: Number of instances
    # N: Number of features
    # C: Number of classes
    # CLASSIFIERCONVERTER: classifier converter corresponding to the op_type
    # a: slope in sigmoid model
    # b: intercept in sigmoid model
    # k: variable in the range [0, C)
    # input: input
    # class_prob_tensor: tensor with class probabilities(function output)
    #
    # Graph:
    #
    #   input [M, N] -> CLASSIFIERCONVERTER -> label [M]
    #                          |
    #                          V
    #                    probability_tensor [M, C]
    #                          |
    #         .----------------'---------.
    #         |                          |
    #         V                          V
    # ARRAYFEATUREEXTRACTOR <- k [1] -> ARRAYFEATUREEXTRACTOR
    #         |                          |
    #         V                          V
    #  transposed_df_col[M, 1] transposed_df_col[M, 1]
    #       |--------------------------|----------.--------------------------.
    #       |                          |          |                          |
    #       |if model.method='sigmoid' |          |if model.method='isotonic'|
    #       |                          |          |                          |
    #       V                          V          |if out_of_bounds='clip'   |
    #      MUL <-------- a -------->  MUL         V                          V
    #       |                          |          CLIP     ...             CLIP
    #       V                          V          |                          |
    #  a_df_prod [M, 1]  ... a_df_prod [M, 1]     V                          V
    #       |                          |  clipped_df [M, 1]...clipped_df [M, 1]
    #       V                          V          |                          |
    #      ADD <--------- b ---------> ADD        '-------------------.------'
    #         |                          |                            |
    #         V                          V                            |
    #  exp_parameter [M, 1] ...   exp_parameter [M, 1]                |
    #         |                          |                            |
    #         V                          V                            |
    #        EXP        ...             EXP                           |
    #         |                          |                            |
    #         V                          V                            |
    #  exp_result [M, 1]  ...    exp_result [M, 1]                    |
    #         |                          |                            |
    #         V                          V                            |
    #       ADD <------- unity -------> ADD                           |
    #         |                          |                            |
    #         V                          V                            |
    #  denominator [M, 1]  ...   denominator [M, 1]                   |
    #         |                          |                            |
    #         V                          V                            |
    #        DIV <------- unity ------> DIV                           |
    #         |                          |                            |
    #         V                          V                            |
    # sigmoid_predict_result [M, 1] ... sigmoid_predict_result [M, 1] |
    #         |                          |                            |
    #         '-----.--------------------'                            |
    #               |-------------------------------------------------'
    #               |
    #               V
    #            CONCAT -> concatenated_prob [M, C]
    #                          |
    #        if  C = 2         |  if C != 2
    #      .-------------------'---------------------------.---------.
    #      |                                               |         |
    #      V                                               |         V
    # ARRAYFEATUREEXTRACTOR <- col_number [1]              |    REDUCESUM
    #                   |                                  |         |
    #                   '--------------------------------. |         |
    # unit_float_tensor [1] -> SUB <- first_col [M, 1] <-' |         |
    #                           |                         /          |
    #                           V                        V           V
    #                         CONCAT                    DIV <- reduced_prob [M]
    #                           |                        |
    #                           V                        |
    #                        class_prob_tensor [M, C] <--'
    model_proba = {RandomForestClassifier}

    if scope.get_options(operator.raw_operator, dict(nocl=False))['nocl']:
        raise RuntimeError(
            "Option 'nocl' is not implemented for operator '{}'.".format(
                operator.raw_operator.__class__.__name__))
    proto_type = guess_proto_type(operator.inputs[0].type)
    proto_type2 = proto_type
    if proto_type2 not in (TensorProto.FLOAT, TensorProto.DOUBLE):
        proto_type2 = TensorProto.FLOAT
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32

    base_model = model.base_estimator
    op_type = sklearn_operator_name_map[type(base_model)]
    n_classes = (len(model.classes_) if hasattr(model, 'classes_') else
                 len(base_model.classes_))
    prob_name = [None] * n_classes

    this_operator = scope.declare_local_operator(op_type, base_model)
    if (container.has_options(base_model, 'raw_scores') and
            not type(base_model) in model_proba):
        container.add_options(id(base_model), {'raw_scores': True})
        scope.add_options(id(base_model), {'raw_scores': True})
    this_operator.inputs = operator.inputs
    label_name = scope.declare_local_variable('label', Int64TensorType())
    df_name = scope.declare_local_variable(
        'uncal_probability', operator.inputs[0].type.__class__())
    this_operator.outputs.append(label_name)
    this_operator.outputs.append(df_name)
    df_inp = df_name.full_name

    for k in range(n_classes):
        cur_k = k
        if n_classes == 2:
            cur_k += 1
        k_name = scope.get_unique_variable_name('k')
        df_col_name = scope.get_unique_variable_name(
            'tdf_col_%d_c%d' % (model_index, k))
        prob_name[k] = scope.get_unique_variable_name(
            'prob_{}_c{}'.format(model_index, k))

        container.add_initializer(k_name, TensorProto.INT64, [], [cur_k])

        container.add_node(
            'ArrayFeatureExtractor', [df_inp, k_name], df_col_name,
            name=scope.get_unique_operator_name(
                'CaliAFE_%d_c%d' % (model_index, k)),
            op_domain='ai.onnx.ml')
        if model.method == 'sigmoid':
            T = _transform_sigmoid(scope, container, model, df_col_name, k,
                                   proto_type)
        else:
            T = _transform_isotonic(scope, container, model, df_col_name,
                                    k, dtype, proto_type)

        prob_name[k] = T
        if n_classes == 2:
            break

    if n_classes == 2:
        zeroth_col_name = scope.get_unique_variable_name(
            'zeroth_col%d' % model_index)
        merged_prob_name = scope.get_unique_variable_name(
            'merged_prob%d' % model_index)
        unit_float_tensor_name = scope.get_unique_variable_name(
            'unit_float_tensor%d' % model_index)

        container.add_initializer(unit_float_tensor_name,
                                  proto_type2, [], [1.0])

        apply_sub(scope, [unit_float_tensor_name, prob_name[0]],
                  zeroth_col_name, container, broadcast=1)
        apply_concat(scope, [zeroth_col_name, prob_name[0]],
                     merged_prob_name, container, axis=1,
                     operator_name=scope.get_unique_variable_name(
                        'CaliConc%d' % model_index))
        class_prob_tensor_name = merged_prob_name
    else:
        concatenated_prob_name = scope.get_unique_variable_name(
            'concatenated_prob')
        reduced_prob_name = scope.get_unique_variable_name('reduced_prob')
        calc_prob_name = scope.get_unique_variable_name('calc_prob')

        apply_concat(scope, prob_name, concatenated_prob_name,
                     container, axis=1)
        if container.target_opset < 13:
            container.add_node(
                'ReduceSum', concatenated_prob_name,
                reduced_prob_name, axes=[1],
                name=scope.get_unique_operator_name('ReduceSum'))
        else:
            axis_name = scope.get_unique_variable_name('axis')
            container.add_initializer(axis_name, TensorProto.INT64, [1], [1])
            container.add_node(
                'ReduceSum', [concatenated_prob_name, axis_name],
                reduced_prob_name,
                name=scope.get_unique_operator_name('ReduceSum'))
        num, deno = _handle_zeros(scope, container, concatenated_prob_name,
                                  reduced_prob_name, n_classes, proto_type)
        apply_div(scope, [num, deno],
                  calc_prob_name, container, broadcast=1,
                  operator_name=scope.get_unique_variable_name(
                    'CaliDiv%d' % model_index))
        class_prob_tensor_name = calc_prob_name
    return class_prob_tensor_name


def convert_sklearn_calibrated_classifier_cv(
        scope: Scope, operator: Operator, container: ModelComponentContainer):
    # Computational graph:
    #
    # In the following graph, variable names are in lower case characters only
    # and operator names are in upper case characters. We borrow operator names
    # from the official ONNX spec:
    # https://github.com/onnx/onnx/blob/master/docs/Operators.md
    # All variables are followed by their shape in [].
    #
    # Symbols:
    # M: Number of instances
    # N: Number of features
    # C: Number of classes
    # CONVERT_BASE_ESTIMATOR: base estimator convert function defined above
    # clf_length: number of calibrated classifiers
    # input: input
    # output: output
    # class_prob: class probabilities
    #
    # Graph:
    #
    #                         input [M, N]
    #                               |
    #           .-------------------|--------------------------.
    #           |                   |                          |
    #           V                   V                          V
    # CONVERT_BASE_ESTIMATOR  CONVERT_BASE_ESTIMATOR ... CONVERT_BASE_ESTIMATOR
    #           |                   |                          |
    #           V                   V                          V
    #  prob_scores_0 [M, C] prob_scores_1 [M, C] ... prob_scores_(clf_length-1)
    #           |                   |                          |  [M, C]
    #           '-------------------|--------------------------'
    #                               V
    #       add_result [M, C] <--- SUM
    #           |
    #           '--> DIV <- clf_length [1]
    #                 |
    #                 V
    #            class_prob [M, C] -> ARGMAX -> argmax_output [M, 1]
    #                                                   |
    #             classes -> ARRAYFEATUREEXTRACTOR  <---'
    #                               |
    #                               V
    #                            output [1]

    op = operator.raw_operator
    classes = op.classes_
    output_shape = (-1,)
    class_type = TensorProto.STRING
    proto_type = guess_proto_type(operator.inputs[0].type)
    proto_type2 = proto_type
    if proto_type2 not in (TensorProto.FLOAT, TensorProto.DOUBLE):
        proto_type2 = TensorProto.FLOAT

    if np.issubdtype(op.classes_.dtype, np.floating):
        class_type = TensorProto.INT32
        classes = classes.astype(np.int32)
    elif np.issubdtype(op.classes_.dtype, np.signedinteger):
        class_type = TensorProto.INT32
    else:
        classes = np.array([s.encode('utf-8') for s in classes])

    clf_length = len(op.calibrated_classifiers_)
    prob_scores_name = []

    clf_length_name = scope.get_unique_variable_name('clf_length')
    classes_name = scope.get_unique_variable_name('classes')
    reshaped_result_name = scope.get_unique_variable_name('reshaped_result')
    argmax_output_name = scope.get_unique_variable_name('argmax_output')
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        'array_feature_extractor_result')
    add_result_name = scope.get_unique_variable_name('add_result')

    container.add_initializer(classes_name, class_type, classes.shape, classes)
    container.add_initializer(clf_length_name, proto_type2,
                              [], [clf_length])

    for clf_index, clf in enumerate(op.calibrated_classifiers_):
        prob_scores_name.append(convert_calibrated_classifier_base_estimator(
            scope, operator, container, clf, clf_index))

    container.add_node('Sum', [s for s in prob_scores_name],
                       add_result_name, op_version=7,
                       name=scope.get_unique_operator_name('Sum'))
    apply_div(scope, [add_result_name, clf_length_name],
              operator.outputs[1].full_name, container, broadcast=1)
    class_prob_name = operator.outputs[1].full_name
    container.add_node('ArgMax', class_prob_name,
                       argmax_output_name,
                       name=scope.get_unique_operator_name('ArgMax'), axis=1)
    container.add_node(
        'ArrayFeatureExtractor', [classes_name, argmax_output_name],
        array_feature_extractor_result_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))

    if class_type == TensorProto.INT32:
        apply_reshape(scope, array_feature_extractor_result_name,
                      reshaped_result_name, container,
                      desired_shape=output_shape)
        apply_cast(scope, reshaped_result_name, operator.outputs[0].full_name,
                   container, to=TensorProto.INT64)
    else:
        apply_reshape(scope, array_feature_extractor_result_name,
                      operator.outputs[0].full_name, container,
                      desired_shape=output_shape)


register_converter('SklearnCalibratedClassifierCV',
                   convert_sklearn_calibrated_classifier_cv,
                   options={'zipmap': [True, False, 'columns'],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import apply_concat, apply_cast
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_concat(scope: Scope, operator: Operator,
                           container: ModelComponentContainer):
    exptype = operator.outputs[0].type
    new_inputs = []
    for inp in operator.inputs:
        if inp.type == exptype:
            new_inputs.append(inp.full_name)
            continue
        name = scope.get_unique_variable_name("{}_cast".format(inp.full_name))
        res = exptype.to_onnx_type()
        et = res.tensor_type.elem_type
        apply_cast(scope, inp.full_name, name, container, to=et)
        new_inputs.append(name)

    apply_concat(scope, new_inputs, operator.outputs[0].full_name,
                 container, axis=1)


register_converter('SklearnConcat', convert_sklearn_concat)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_normalizer
from ..common.data_types import DoubleTensorType
from .common import concatenate_variables


def convert_sklearn_normalizer(scope: Scope, operator: Operator,
                               container: ModelComponentContainer):
    if len(operator.inputs) > 1:
        # If there are multiple input tensors,
        # we combine them using a FeatureVectorizer
        feature_name = concatenate_variables(scope, operator.inputs, container)
    else:
        # No concatenation is needed, we just use the first variable's name
        feature_name = operator.inputs[0].full_name
    op = operator.raw_operator
    norm_map = {'max': 'MAX', 'l1': 'L1', 'l2': 'L2'}
    if op.norm in norm_map:
        norm = norm_map[op.norm]
    else:
        raise RuntimeError("Invalid norm '%s'. You may raise an issue"
                           "at https://github.com/onnx/sklearn-onnx/"
                           "issues." % op.norm)
    use_float = type(operator.inputs[0].type) not in (DoubleTensorType, )
    apply_normalizer(
        scope, feature_name, operator.outputs[0].full_name, container,
        norm=norm, use_float=use_float)


register_converter('SklearnNormalizer', convert_sklearn_normalizer)



# SPDX-License-Identifier: Apache-2.0


from onnx.helper import make_tensor
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_mul
from ..common.utils_classifier import _finalize_converter_classes
from ..common.data_types import guess_proto_type, Int64TensorType
from .._supported_operators import sklearn_operator_name_map
from ..proto import onnx_proto


def convert_voting_classifier(scope: Scope, operator: Operator,
                              container: ModelComponentContainer):
    """
    Converts a *VotingClassifier* into *ONNX* format.

    *predict_proba* is not defined by *scikit-learn* when ``voting='hard'``.
    The converted model still defines a probability vector equal to the
    highest probability obtained for each class over all estimators.

    *scikit-learn* enables both modes, transformer and predictor
    for the voting classifier. *ONNX* does not make this
    distinction and always creates two outputs, labels
    and probabilities.
    """
    if scope.get_options(operator.raw_operator, dict(nocl=False))['nocl']:
        raise RuntimeError(
            "Option 'nocl' is not implemented for operator '{}'.".format(
                operator.raw_operator.__class__.__name__))
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator
    n_classes = len(op.classes_)

    classes_ind_name = scope.get_unique_variable_name('classes_ind')
    container.add_initializer(classes_ind_name, onnx_proto.TensorProto.INT64,
                              (1, n_classes), list(range(n_classes)))

    probs_names = []
    one_name = None
    for i, estimator in enumerate(op.estimators_):
        if estimator is None:
            continue

        op_type = sklearn_operator_name_map[type(estimator)]

        this_operator = scope.declare_local_operator(op_type, estimator)
        this_operator.inputs = operator.inputs

        label_name = scope.declare_local_variable(
            'label_%d' % i, Int64TensorType())
        prob_name = scope.declare_local_variable(
            'voting_proba_%d' % i, operator.inputs[0].type.__class__())
        this_operator.outputs.append(label_name)
        this_operator.outputs.append(prob_name)

        if op.voting == 'hard':
            if one_name is None:
                shape_name = scope.get_unique_variable_name('shape')
                container.add_node(
                    'Shape', prob_name.onnx_name, shape_name,
                    name=scope.get_unique_operator_name('Shape'))
                zero_name = scope.get_unique_variable_name('zero')
                container.add_node(
                    'ConstantOfShape', shape_name, zero_name,
                    name=scope.get_unique_operator_name('CoSA'),
                    value=make_tensor("value", proto_dtype,
                                      (1, ), [0.]), op_version=9)
                one_name = scope.get_unique_variable_name('one')
                container.add_node(
                    'ConstantOfShape', shape_name, one_name,
                    name=scope.get_unique_operator_name('CoSB'),
                    value=make_tensor("value", proto_dtype,
                                      (1, ), [1.]), op_version=9)

            argmax_output_name = scope.get_unique_variable_name(
                'argmax_output')
            container.add_node('ArgMax', prob_name.onnx_name,
                               argmax_output_name,
                               name=scope.get_unique_operator_name('ArgMax'),
                               axis=1)

            equal_name = scope.get_unique_variable_name('equal')
            container.add_node('Equal', [argmax_output_name, classes_ind_name],
                               equal_name,
                               name=scope.get_unique_operator_name('Equal'))

            max_proba_name = scope.get_unique_variable_name('probsmax')
            container.add_node('Where', [equal_name, one_name, zero_name],
                               max_proba_name,
                               name=scope.get_unique_operator_name('Where'))
            prob_name = max_proba_name
        else:
            prob_name = prob_name.onnx_name

        if op.weights is not None:
            val = op.weights[i] / op.weights.sum()
        else:
            val = 1. / len(op.estimators_)

        weights_name = scope.get_unique_variable_name('w%d' % i)
        container.add_initializer(
            weights_name, proto_dtype, [1], [val])
        wprob_name = scope.get_unique_variable_name('wprob_name')
        apply_mul(scope, [prob_name, weights_name],
                  wprob_name, container, broadcast=1)
        probs_names.append(wprob_name)

    if op.flatten_transform in (False, None):
        container.add_node('Sum', probs_names,
                           operator.outputs[1].full_name,
                           name=scope.get_unique_operator_name('Sum'))
    else:
        raise NotImplementedError(
            "flatten_transform==True is not implemented yet. "
            "You may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues.")

    # labels
    label_name = scope.get_unique_variable_name('label_name')
    container.add_node('ArgMax', operator.outputs[1].full_name, label_name,
                       name=scope.get_unique_operator_name('ArgMax'), axis=1)
    _finalize_converter_classes(scope, label_name,
                                operator.outputs[0].full_name, container,
                                op.classes_, proto_dtype)


register_converter('SklearnVotingClassifier',
                   convert_voting_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import (
    Int64TensorType, guess_numpy_type, guess_proto_type)
from ..algebra.onnx_ops import (
    OnnxAdd, OnnxCast, OnnxDiv, OnnxMatMul, OnnxSub)


def convert_pls_regression(scope: Scope, operator: Operator,
                           container: ModelComponentContainer):
    X = operator.inputs[0]
    op = operator.raw_operator
    opv = container.target_opset
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    if type(X.type) == Int64TensorType:
        X = OnnxCast(X, to=proto_dtype, op_version=opv)

    coefs = op.x_mean_ if hasattr(op, 'x_mean_') else op._x_mean
    std = op.x_std_ if hasattr(op, 'x_std_') else op._x_std
    ym = op.y_mean_ if hasattr(op, 'x_mean_') else op._y_mean

    norm_x = OnnxDiv(
        OnnxSub(X, coefs.astype(dtype), op_version=opv),
        std.astype(dtype), op_version=opv)
    dot = OnnxMatMul(norm_x, op.coef_.astype(dtype),
                     op_version=opv)
    pred = OnnxAdd(dot, ym.astype(dtype),
                   op_version=opv, output_names=operator.outputs)
    pred.add_to(scope, container)


register_converter('SklearnPLSRegression', convert_pls_regression)



"""
@file
@brief Overloads :epkg:`TfidfVectorizer` and :epkg:`CountVectorizer`.
"""
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
try:
    from sklearn.feature_extraction.text import (
        _VectorizerMixin as VectorizerMixin)
except ImportError:  # pragma: no cover
    # scikit-learn < 0.23
    from sklearn.feature_extraction.text import VectorizerMixin


class NGramsMixin(VectorizerMixin):
    """
    Overloads method `_word_ngrams
    <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/feature_extraction/text.py#L148>`_
    to get tuples instead of string in member `vocabulary_
    <https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.CountVectorizer.html>`_.
    of `TfidfVectorizer` or :epkg:`CountVectorizer`.
    It contains the list of n-grams used to process documents.
    See :class:`TraceableCountVectorizer` and :class:`TraceableTfidfVectorizer`
    for example.
    """

    def _word_ngrams(self, tokens, stop_words=None):
        """Turn tokens into a sequence of n-grams after stop words filtering"""
        # handle stop words
        if tokens is not None:
            new_tokens = []
            for token in tokens:
                new_tokens.append(
                    (token,) if isinstance(token, str) else token)
            tokens = new_tokens

        if stop_words is not None:
            tokens = [(w, ) for w in tokens if w not in stop_words]

        # handle token n-grams
        min_n, max_n = self.ngram_range
        if max_n != 1:
            original_tokens = tokens
            if min_n == 1:
                # no need to do any slicing for unigrams
                # just iterate through the original tokens
                tokens = list(original_tokens)
                min_n += 1
            else:
                tokens = []

            n_original_tokens = len(original_tokens)

            # bind method outside of loop to reduce overhead
            tokens_append = tokens.append

            def space_join(tokens):
                new_tokens = []
                for token in tokens:
                    if isinstance(token, str):
                        new_tokens.append(token)
                    elif isinstance(token, tuple):
                        new_tokens.extend(token)
                    else:
                        raise TypeError(  # pragma: no cover
                            f"Unable to build a n-grams out of {tokens}.")
                return tuple(new_tokens)

            for n in range(min_n,
                           min(max_n + 1, n_original_tokens + 1)):
                for i in range(n_original_tokens - n + 1):
                    tokens_append(space_join(original_tokens[i: i + n]))
        return tokens


class TraceableCountVectorizer(CountVectorizer, NGramsMixin):
    """
    Inherits from :class:`NGramsMixin` which overloads method `_word_ngrams
    <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/feature_extraction/text.py#L148>`_
    to keep more information about n-grams but still produces the same
    outputs than `CountVectorizer`.

    .. runpython::
        :showcode:

        import numpy
        from sklearn.feature_extraction.text import CountVectorizer
        from mlinsights.mlmodel.sklearn_text import TraceableCountVectorizer
        from pprint import pformat

        corpus = numpy.array([
            "This is the first document.",
            "This document is the second document.",
            "Is this the first document?",
            "",
        ]).reshape((4, ))

        print('CountVectorizer from scikit-learn')
        mod1 = CountVectorizer(ngram_range=(1, 2))
        mod1.fit(corpus)
        print(mod1.transform(corpus).todense()[:2])
        print(pformat(mod1.vocabulary_)[:100])

        print('TraceableCountVectorizer from scikit-learn')
        mod2 = TraceableCountVectorizer(ngram_range=(1, 2))
        mod2.fit(corpus)
        print(mod2.transform(corpus).todense()[:2])
        print(pformat(mod2.vocabulary_)[:100])

    A weirder example with
    @see cl TraceableTfidfVectorizer shows more differences.
    """

    def _word_ngrams(self, tokens, stop_words=None):
        return NGramsMixin._word_ngrams(
            self, tokens=tokens, stop_words=stop_words)


class TraceableTfidfVectorizer(TfidfVectorizer, NGramsMixin):
    """
    Inherits from :class:`NGramsMixin` which overloads method `_word_ngrams
    <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/feature_extraction/text.py#L148>`_
    to keep more information about n-grams but still produces the same
    outputs than `TfidfVectorizer`.

    .. runpython::
        :showcode:

        import numpy
        from sklearn.feature_extraction.text import TfidfVectorizer
        from mlinsights.mlmodel.sklearn_text import TraceableTfidfVectorizer
        from pprint import pformat

        corpus = numpy.array([
            "This is the first document.",
            "This document is the second document.",
            "Is this the first document?",
            "",
        ]).reshape((4, ))

        print('TfidfVectorizer from scikit-learn')
        mod1 = TfidfVectorizer(ngram_range=(1, 2),
                               token_pattern="[a-zA-Z ]{1,4}")
        mod1.fit(corpus)
        print(mod1.transform(corpus).todense()[:2])
        print(pformat(mod1.vocabulary_)[:100])

        print('TraceableTfidfVectorizer from scikit-learn')
        mod2 = TraceableTfidfVectorizer(ngram_range=(1, 2),
                                       token_pattern="[a-zA-Z ]{1,4}")
        mod2.fit(corpus)
        print(mod2.transform(corpus).todense()[:2])
        print(pformat(mod2.vocabulary_)[:100])
    """

    def _word_ngrams(self, tokens, stop_words=None):
        return NGramsMixin._word_ngrams(
            self, tokens=tokens, stop_words=stop_words)
